from pathlib import Path

import networkx as nx
import numpy as np
import osmnx
from geopandas import GeoDataFrame

from ra2ce.analysis.analysis_config_data.analysis_config_data import (
    AnalysisSectionLosses,
)
from ra2ce.analysis.analysis_input_wrapper import AnalysisInputWrapper
from ra2ce.analysis.losses.analysis_losses_protocol import AnalysisLossesProtocol
from ra2ce.analysis.losses.weighing_analysis.weighing_analysis_factory import (
    WeighingAnalysisFactory,
)
from ra2ce.network.graph_files.graph_file import GraphFile
from ra2ce.network.hazard.hazard_names import HazardNames


class SingleLinkRedundancy(AnalysisLossesProtocol):
    analysis: AnalysisSectionLosses
    graph_file: GraphFile
    input_path: Path
    static_path: Path
    output_path: Path
    hazard_names: HazardNames

    def __init__(self, analysis_input: AnalysisInputWrapper) -> None:
        self.analysis = analysis_input.analysis
        self.graph_file = analysis_input.graph_file
        self.input_path = analysis_input.input_path
        self.static_path = analysis_input.static_path
        self.output_path = analysis_input.output_path
        self.hazard_names = analysis_input.hazard_names
        self.result = None

    def execute(self) -> GeoDataFrame:
        """This is the function to analyse roads with a single link disruption and an alternative route."""
        # TODO adjust to the right names of the RA2CE tool
        # if 'road_usage_data_path' in InputDict:
        #     road_usage_data = pd.read_excel(InputDict.road_usage_data_path)
        #     road_usage_data.dropna(axis=0, how='all', subset=['vehicle_type'], inplace=True)
        #     aadt_names = [aadt_name for aadt_name in road_usage_data['attribute_name'] if aadt_name == aadt_name]
        # else:
        #     aadt_names = None
        #     road_usage_data = pd.DataFrame()

        # create a geodataframe from the graph
        _gdf_graph = osmnx.graph_to_gdfs(self.graph_file.get_graph(), nodes=False)

        # list for the length of the alternative routes
        _current_value_list = []
        _alt_value_list = []
        _alt_nodes_list = []
        _diff_value_list = []
        _detour_exist_list = []

        _weighing_analyser = WeighingAnalysisFactory.get_analysis(
            self.analysis.weighing
        )

        # Loop over all edges to temporarily remove them and calculate the alternative route
        for e_remove in list(self.graph_file.graph.edges.data(keys=True)):
            u, v, k, _weighing_analyser.edge_data = e_remove
            _current_value = _weighing_analyser.get_current_value()

            # remove the edge
            self.graph_file.graph.remove_edge(u, v, k)

            _alt_value, _alt_nodes, _connected, _diff = np.nan, np.nan, 0, np.nan
            if nx.has_path(self.graph_file.graph, u, v):

                # calculate the alternative distance/time and path if that edge is unavailable
                [_alt_value, _alt_nodes] = nx.single_source_dijkstra(
                    self.graph_file.graph,
                    u,
                    v,
                    weight=self.analysis.weighing.config_value,
                )

                # calculate the difference in distance
                _diff = round(_alt_value - _current_value, 3)

                _connected = 1

            _current_value_list.append(_current_value)
            _alt_value_list.append(_alt_value)
            _alt_nodes_list.append(_alt_nodes)
            _diff_value_list.append(_diff)
            _detour_exist_list.append(_connected)

            # add edge again to the graph
            self.graph_file.graph.add_edge(u, v, k, **_weighing_analyser.edge_data)

        # Add the updated/new columns to the geodataframe
        _gdf_graph[self.analysis.weighing.config_value] = _current_value_list
        _gdf_graph[f"alt_{self.analysis.weighing.config_value}"] = _alt_value_list
        _gdf_graph["alt_nodes"] = _alt_nodes_list
        _gdf_graph[f"diff_{self.analysis.weighing.config_value}"] = _diff_value_list
        _gdf_graph["detour"] = _detour_exist_list

        # Extra calculation possible (like multiplying the disruption time with the cost for disruption)
        # todo: input here this option

        return _gdf_graph



"""
                    GNU GENERAL PUBLIC LICENSE
                      Version 3, 29 June 2007

    Risk Assessment and Adaptation for Critical Infrastructure (RA2CE).
    Copyright (C) 2023 Stichting Deltares

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

from pathlib import Path

import geopandas as gpd
import networkx as nx

from ra2ce.network.exporters.geodataframe_network_exporter import (
    GeoDataFrameNetworkExporter,
)
from ra2ce.network.exporters.multi_graph_network_exporter import (
    MultiGraphNetworkExporter,
)
from ra2ce.network.exporters.network_exporter_base import (
    NETWORK_TYPE,
    NetworkExporterBase,
)


class NetworkExporterFactory:
    _exporter: NetworkExporterBase

    def export(
        self,
        network: NETWORK_TYPE,
        basename: str,
        output_dir: Path,
        export_types: list[str],
    ) -> None:
        _exporter_type = self.get_exporter_type(network)
        self._exporter = _exporter_type(basename, export_types)
        self._exporter.export(output_dir, network)

    def get_pickle_path(self) -> Path:
        return self._exporter.pickle_path

    @staticmethod
    def get_exporter_type(network: NETWORK_TYPE) -> type[NetworkExporterBase]:
        _network_type = type(network)
        if _network_type == gpd.GeoDataFrame:
            return GeoDataFrameNetworkExporter
        if _network_type in [
            nx.MultiGraph,
            nx.MultiDiGraph,
        ]:
            return MultiGraphNetworkExporter
        raise ValueError(f"Unsupported network type: {_network_type}")



"""
                    GNU GENERAL PUBLIC LICENSE
                      Version 3, 29 June 2007

    Risk Assessment and Adaptation for Critical Infrastructure (RA2CE).
    Copyright (C) 2023 Stichting Deltares

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

from pathlib import Path

import geopandas as gpd
import pandas as pd

from ra2ce.analysis.losses.traffic_analysis.equity_analysis import EquityAnalysis
from ra2ce.analysis.losses.traffic_analysis.traffic_analysis import TrafficAnalysis
from ra2ce.analysis.losses.traffic_analysis.traffic_analysis_base import (
    TrafficAnalysisBase,
)


class TrafficAnalysisFactory:
    @staticmethod
    def get_analysis(
        road_network: gpd.GeoDataFrame,
        od_table: gpd.GeoDataFrame,
        destination_names: str,
        equity_data: pd.DataFrame,
    ) -> TrafficAnalysisBase:
        """
        Gets an instance of an `TrafficAnalysisBase` based on whether `equity_data` is provided or not.

        Args:
            road_network (gpd.GeoDataFrame): General dataframe for the road network to analyze.
            od_table (gpd.GeoDataFrame): Origins - destination table dataframe.
            destination_names (str): Destination names.
            equity_data (pd.DataFrame): Dataframe contaning region - weight information.

        Returns:
            TrafficAnalysisBase: Object to make an equity analysis.
        """
        if isinstance(equity_data, pd.DataFrame) and not equity_data.empty:
            return EquityAnalysis(
                road_network, od_table, destination_names, equity_data
            )
        return TrafficAnalysis(road_network, od_table, destination_names)

    @staticmethod
    def read_equity_weights(equity_weight_file: Path) -> pd.DataFrame:
        """
        Reads the equity data from a geojson fileand loads it into a pandas dataframe.

        Args:
            equity_weight_file (Path): File containing values of region and weight.

        Returns:
            pd.DataFrame: Dataframe representing the geojson data.
        """
        if not equity_weight_file or not equity_weight_file.exists():
            return pd.DataFrame()

        _separator = (
            ";" if ";" in equity_weight_file.read_text().splitlines()[0] else ","
        )
        return pd.read_csv(equity_weight_file, sep=_separator)



"""
                    GNU GENERAL PUBLIC LICENSE
                      Version 3, 29 June 2007

    Risk Assessment and Adaptation for Critical Infrastructure (RA2CE).
    Copyright (C) 2023 Stichting Deltares

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""


import logging
from pathlib import Path

import pandas as pd


class DamageFractionUniform:
    """
    Uniform: assuming the same curve for
    each road type and lane numbers and any other metadata


    self.raw_data (pd.DataFrame) : Raw data from the csv file
    self.data (pd.DataFrame) : index = hazard severity (e.g. flood depth); column 0 = damage fraction

    """

    def __init__(self, name=None, hazard_unit=None):
        self.name = name
        self.hazard_unit = hazard_unit
        self.interpolator = None

    def from_csv(self, path: Path, sep=",") -> None:
        """Construct object from csv file. Damage curve name is inferred from filename

        Arguments:
            *path* (Path) : Path to the csv file
            *sep* (str) : csv seperator
            *output_unit* (str) : desired output unit (default = 'm')

        The CSV file should have the following structure:
         - column 1: hazard severity
         - column 2: damage fraction
         - row 1: column names
         - row 2: unit of column:

        Example:
                +- ------+-------------------------------+
                | depth | damage                        |
                +-------+-------------------------------+
                | cm    | % of total construction costs |
                +-------+-------------------------------+
                | 0     | 0                             |
                +-------+-------------------------------+
                | 50    | 0.25                          |
                +-------+-------------------------------+
                | 100   | 0.42                          |
                +-------+-------------------------------+


        """
        self.name = path.stem
        self.raw_data = pd.read_csv(path, index_col=0, sep=sep)
        self.origin_path = path  # to track the original path from which the object was constructed; maybe also date?

        # identify unit and drop from data
        self.hazard_unit = self.raw_data.index[0]
        self.data = self.raw_data.drop(
            self.hazard_unit
        )  # Todo: This could also be a series instead of DataFrame

        # convert data to floats
        self.data = self.data.astype("float")
        self.data.index = self.data.index.astype("float")

        self.convert_hazard_severity_unit()

    def convert_hazard_severity_unit(self, desired_unit="m") -> None:
        """Converts hazard severity values to a different unit
        Arguments:
            self.hazard_unit - implicit (string)
            *desired_unit* (string)

        Effect: converts the values in self.data; and sets the new damage_unit to the desired unit

        """
        if desired_unit == self.hazard_unit:
            logging.info(
                "Damage units are already in the desired format {}".format(desired_unit)
            )
            return None

        if self.hazard_unit == "cm" and desired_unit == "m":
            scaling_factor = 1 / 100
            self.data.index = self.data.index * scaling_factor
            logging.info(
                "Hazard severity from {} data was scaled by a factor {}, to convert from {} to {}".format(
                    self.origin_path, scaling_factor, self.hazard_unit, desired_unit
                )
            )
            self.damage_unit = desired_unit
            return None
        else:
            logging.warning(
                "Hazard severity scaling from {} to {} is not  supported".format(
                    self.hazard_unit, desired_unit
                )
            )
            return None

    def create_interpolator(self):
        """Create interpolator object from loaded data
        sets result to self.interpolator (Scipy interp1d)
        """
        from scipy.interpolate import interp1d

        x_values = self.data.index.values
        y_values = self.data.values[:, 0]

        self.interpolator = interp1d(
            x=x_values,
            y=y_values,
            fill_value=(
                y_values[0],
                y_values[-1],
            ),  # fraction damage (y) if hazard severity (x) is outside curve range
            bounds_error=False,
        )

        return None

    def __repr__(self):
        if self.interpolator:
            string = (
                "DamageFractionUniform with name: "
                + self.name
                + " interpolator: {}".format(
                    list(zip(self.interpolator.y, self.interpolator.x))
                )
            )
        else:
            string = "DamageFractionUniform with name: " + self.name
        return string



"""
                    GNU GENERAL PUBLIC LICENSE
                      Version 3, 29 June 2007

    Risk Assessment and Adaptation for Critical Infrastructure (RA2CE).
    Copyright (C) 2023 Stichting Deltares

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

import logging
from abc import ABC, abstractmethod

import numpy as np
import pandas as pd
from geopandas import GeoDataFrame
from scipy.interpolate import interp1d

from ra2ce.analysis.analysis_config_data.enums.damage_curve_enum import DamageCurveEnum
from ra2ce.analysis.damages.damages_lookup import LookUp as lookup
from ra2ce.analysis.damages.damages_lookup import dataframe_lookup
from ra2ce.analysis.damages.damages_utils import (
    clean_lane_data,
    create_summary_statistics,
    scale_damage_using_lanes,
)


class DamageNetworkBase(ABC):
    """A road network gdf with hazard data stored in it, and for which damages can be calculated"""

    def __init__(
        self,
        road_gdf: GeoDataFrame,
        val_cols: list[str],
        representative_damage_percentage: float,
    ):
        """Construct the Data"""
        self.val_cols = val_cols
        self.gdf = road_gdf
        # set of hazard info per event
        self.stats = set([x.split("_")[-1] for x in val_cols])
        self.representative_damage_percentage = representative_damage_percentage
        # TODO: also track the damage cols after the dam calculation, that is useful for the risk calc. module
        # TODO: also create constructors of the children of this class

    @abstractmethod
    def main(self, damage_function: DamageCurveEnum, manual_damage_functions):
        """
        Controller for doing the EAD calculation

        Args:
            damage_function (DamageCurveEnum): damage function key name that is to be used
            manual_damage_functions (ManualDamageFunctions): `ManualDamageFunctions` object
        """
        raise ValueError("Needs to be implemented in concrete child class.")

    # events is missing
    def do_cleanup_and_mask_creation(self):
        """Call all cleanup and mask functions, because this is a standard procedure in most calculations"""
        # CLEANUPS (on the original gdf)
        self.remap_road_types_to_fewer_classes()
        self.clean_and_interpolate_missing_lane_data()
        self.fix_extraordinary_lanes()

        # Mask creation
        self.create_mask()
        self.remove_unclassified_road_types_from_mask()

    ### Generic cleanup functionality
    def fix_extraordinary_lanes(self):
        """Remove exceptionally high/low lane numbers in self.gdf"""
        # fixing lanes
        df = self.gdf
        df["lanes_copy"] = df["lanes"].copy()
        df["lanes"] = df["lanes_copy"].where(
            df["lanes_copy"].astype(float) >= 1.0, other=1.0
        )
        df["lanes"] = df["lanes_copy"].where(
            df["lanes_copy"].astype(float) <= 6.0, other=6.0
        )
        df = self.gdf

    def clean_and_interpolate_missing_lane_data(self):
        # cleanup and complete the lane data.
        ### Try to convert all data to floats
        try:
            self.gdf.lanes = self.gdf.lanes.astype(
                "float"
            )  # floats instead of ints because ints cannot be nan.
        except Exception:
            logging.warning(
                "Available lane data cannot simply be converted to float/int, RA2CE will try a clean-up."
            )
            self.gdf.lanes = clean_lane_data(self.gdf.lanes)

        # round to nearest integer, but save as float format
        self.gdf.lanes = self.gdf.lanes.round(0)

        # boolean with trues for all nans, i.e. all road segements without lane data
        nans = self.gdf.lanes.isnull()
        if nans.sum() > 0:
            logging.warning(
                """Of the {} road segments, only {} had lane data, so for {} the '
                                    lane data will be interpolated from the existing data""".format(
                    len(self.gdf.lanes), (~nans).sum(), nans.sum()
                )
            )
            lane_stats = create_summary_statistics(self.gdf)

            # Replace the missing lane data the neat way (without pandas SettingWithCopyWarning)
            lane_nans_mask = self.gdf.lanes.isnull()
            self.gdf.loc[lane_nans_mask, "lanes"] = self.gdf.loc[
                lane_nans_mask, "road_type"
            ].replace(lane_stats)
            logging.warning(
                "Interpolated the missing lane data as follows: {}".format(lane_stats)
            )
            # all nans should be replaced
            assert np.nan not in self.gdf.lanes.unique()

        # TODO: think about if this is the best option
        self.gdf.loc[self.gdf["lanes"] == 0, "lanes"] = 1

    def remap_road_types_to_fewer_classes(self):
        """
        Creates a new new column road_types, which has a fewer number of road type categories
        e.g. -> 'motorway_junction' -> 'motorway'
        (Renames highway column to infra_type)
        :return:
        """
        # reduce the number of road types (col 'infra_type') to smaller number of road_types for which damage curves exist
        road_mapping_dict = (
            lookup.road_mapping()
        )  # The lookup class contains all kinds of data
        gdf = self.gdf
        gdf.rename(
            columns={"highway": "infra_type"}, inplace=True
        )  # Todo: this should probably not be done here
        gdf["road_type"] = gdf["infra_type"]
        gdf = gdf.replace({"road_type": road_mapping_dict})
        self.gdf = gdf

    def remove_unclassified_road_types_from_mask(self):
        """
        Drop all rows with road types classified as 'none' from the self._gdf_mask

        :return:
        """
        df = self._gdf_mask
        if "none" in df["road_type"].unique():
            to_drop = df.loc[df["road_type"] == "none"]
            logging.warning(
                "We will drop {} rows for which the road_type is unrecognized".format(
                    to_drop.shape[0]
                )
            )
            self._gdf_mask = df.loc[~(df["road_type"] == "none")]

    ### Damage handlers
    def calculate_damage_manual_functions(self, events, manual_damage_functions):
        """
        Arguments:
            *events* (list) : list of events (or return periods) to iterate over, these should match the hazard column names
            *manual_damage_functions* (RA2CE ManualDamageFunctions object) :
        """

        # Todo: Dirty fixes, these should be read from the init
        hazard_prefix = "F"

        # dataframe to carry out the damage calculation #todo: this is a bit dirty
        df = self._gdf_mask

        assert manual_damage_functions is not None, "No damage functions were loaded"

        for _loaded_func in manual_damage_functions.loaded:
            # Add max damage values to df
            df = _loaded_func.add_max_damage(df, _loaded_func.prefix)
            for event in events:
                # Add apply interpolator objects
                event_prefix = event
                df = _loaded_func.calculate_damage(
                    df, _loaded_func.prefix, hazard_prefix, event_prefix
                )

        # Only transfer the final results to the damage column
        dam_cols = [c for c in df.columns if c.startswith("dam_")]
        self.gdf[dam_cols] = df[dam_cols]
        logging.info(
            "Damage calculation with the manual damage functions was succesfull."
        )

    def calculate_damage_HZ(self, events):
        """
        Arguments:
           *events* (list) = list of events (or return periods) to iterate over, these should match the hazard column names
        """
        # These factors are derived from: Van Ginkel et al. 2021: https://nhess.copernicus.org/articles/21/1011/2021/
        logging.warning(
            "Damage calculations with Huizinga curves are based on Van Ginkel et al. 2021: https://nhess.copernicus.org/articles/21/1011/2021/"
        )
        logging.warning(
            """All damages represent the former EU-28 (before Brexit), 2015-pricelevel in Euro's.
                            To convert to local currency, these need to be:
                                multiplied by the ratio (pricelevel_XXXX / pricelevel_2015)
                                multiply by the ratio (local_GDP_per_capita / EU-28-2015-GDP_per_capita)          
                            EU-28-2015-GDP_per_capita = 39.200 euro
                        """
        )
        logging.warning(
            "These numbers assume that motorways that each driving direction is mapped as a seperate segment such as in OSM!!!"
        )

        # Todo: Dirty fixes, these should be read from the init
        hazard_prefix = "F"
        end = "me"  # indicate that you want to use the mean

        # Load the Huizinga damage functions
        curve_name = "HZ"

        df_max_damages_huizinga = pd.DataFrame.from_dict(
            lookup.get_max_damages_huizinga()
        )
        interpolator = lookup.get_flood_curves()[
            "HZ"
        ]  # input: water depth (cm); output: damage (fraction road construction costs)

        df = self._gdf_mask
        df["lanes"] = df["lanes"].astype(int)
        df["max_dam_hz"] = df.apply(
            dataframe_lookup,
            args=(df_max_damages_huizinga, ["lanes", "road_type"]),
            axis=1,
        )

        for event in events:
            df["dam_{}_{}".format(event, curve_name)] = round(
                df["max_dam_hz"].astype(float)  # max damage (euro/m)
                * interpolator(df["{}_{}_{}".format(hazard_prefix, event, end)]).astype(
                    float
                )  # damage curve  (-)
                * df["{}_{}_{}".format(hazard_prefix, event, "fr")].astype(
                    float
                )  # inundated fraction (-)
                * df["length"],
                2,
            )  # length segment (m)

        # Add the new columns add the right location to the df
        dam_cols = [c for c in df.columns if c.startswith("dam_")]

        self.gdf[dam_cols] = df[dam_cols]
        logging.info(
            "calculate_damage_HZ(): Damage calculation with the Huizinga damage functions was successful"
        )

    def calculate_damage_OSdaMage(self, events):
        """Damage calculation with the OSdaMage functions"""

        def interpolate_damage(row, representative_damage_percentage):
            # Extract the tuple of damage values from the row
            damage_values = row["dam_{}_{}_quartiles".format(curve_name, event)]

            # Quantile values corresponding to the damage values
            percentages = [0, 25, 50, 75, 100]

            # Perform linear interpolation using interp1d from scipy
            _interpolator = interp1d(
                percentages, damage_values, kind="linear", fill_value="extrapolate"
            )

            # Interpolate the damage value for the given representative_damage_percentage
            interpolated_damage = _interpolator(representative_damage_percentage)

            return interpolated_damage

        # These factors are derived from: Van Ginkel et al. 2021: https://nhess.copernicus.org/articles/21/1011/2021/
        logging.warning(
            """Damage calculations with OSdaMage functions are based on 
            Van Ginkel et al. 2021: https://nhess.copernicus.org/articles/21/1011/2021/"""
        )
        logging.warning(
            """All damages represent the former EU-28 (before Brexit), 2015-pricelevel in Euro's.
                            To convert to local currency, these need to be:
                                multiplied by the ratio (pricelevel_XXXX / pricelevel_2015)
                                multiply by the ratio (local_GDP_per_capita / EU-28-2015-GDP_per_capita)          
                            EU-28-2015-GDP_per_capita = 39.200 euro
                        """
        )
        logging.warning(
            "These numbers assume that motorways that each driving direction is mapped as a seperate segment such as in OSM!!!"
        )

        # Todo: Dirty fixes, these should be read from the configuration file
        hazard_prefix = "F"
        end = "me"  # indicate that you want to use the mean

        # Load the OSdaMage functions
        max_damages = lookup.get_max_damages_osd()
        interpolators = lookup.get_flood_curves()
        interpolators.pop(
            "HZ"
        )  # input: water depth (cm); output: damage (fraction road construction costs)
        lane_scale_factors = lookup.get_lane_number_damage_correction()

        # Prepare the output files
        df = self._gdf_mask
        df["tuple"] = [tuple([0] * 5)] * len(df["lanes"])

        # CALCULATE MINIMUM AND MAXIMUM CONSTRUCTION COST PER ROAD TYPE
        # pre-calculation of max damages per percentage (same for each C1-C6 category)
        df["lower_damage"] = (
            df["road_type"].copy().map(max_damages["Lower"])
        )  # i.e. min construction costs
        df["upper_damage"] = (
            df["road_type"].copy().map(max_damages["Upper"])
        )  # i.e. max construction costs

        # apply damage correction for lanes
        cols_to_scale = ["lower_damage", "upper_damage"]
        df = scale_damage_using_lanes(lane_scale_factors, df, cols_to_scale)

        # create separate column for each percentage of construction costs (is faster then tuple)
        for percentage in [
            0,
            25,
            50,
            75,
            100,
        ]:  # So this interpolates the min to the max damage
            df["damage_{}".format(percentage)] = (
                df["upper_damage"] * percentage / 100
            ) + (df["lower_damage"] * (100 - percentage) / 100)

        for curve_name, interpolator in interpolators.items():
            # print(curve_name, interpolator)
            for event in events:
                for percentage in [0, 25, 50, 75, 100]:
                    df["dam_{}_{}_{}".format(percentage, curve_name, event)] = round(
                        df["damage_{}".format(percentage)].astype(
                            float
                        )  # max damage (in euro/m)
                        * interpolator(
                            df["{}_{}_{}".format(hazard_prefix, event, end)]
                        ).astype(
                            float
                        )  # damage curve: fraction f(depth-cm) #Todo check units
                        * df["{}_{}_{}".format(hazard_prefix, event, "fr")].astype(
                            float
                        )  # inundated fraction of the segment should be in km. because max damage (in euro/km)
                        * df["length"].astype(float),
                        3,
                    )

                # This wraps it all in tuple again
                df["dam_{}_{}_quartiles".format(curve_name, event)] = tuple(
                    zip(
                        df["dam_0_{}_{}".format(curve_name, event)],
                        df["dam_25_{}_{}".format(curve_name, event)],
                        df["dam_50_{}_{}".format(curve_name, event)],
                        df["dam_75_{}_{}".format(curve_name, event)],
                        df["dam_100_{}_{}".format(curve_name, event)],
                    )
                )
                df[f"dam_{curve_name}_{event}_representative"] = (
                    df.apply(
                        lambda row: interpolate_damage(
                            row, self.representative_damage_percentage
                        ),
                        axis=1,
                    )
                ).astype(float)

                # And throw way all intermediate results (that are not in the tuple)
                df = df.drop(
                    columns=[
                        "dam_{}_{}_{}".format(percentage, curve_name, event)
                        for percentage in [0, 25, 50, 75, 100]
                    ]
                )

        df = df.drop(columns=[c for c in df.columns if c.startswith("damage_")])

        # drop invalid combinations of damage curves and road types (C1-C4 for motorways; C5,C6 for other)
        all_dam_cols = [c for c in df.columns if c.startswith("dam_")]
        motorway_curves = [
            c for c in all_dam_cols if int(c.split("_")[1][-1]) <= 4
        ]  # C1-C4
        other_curves = [
            c for c in all_dam_cols if int(c.split("_")[1][-1]) > 4
        ]  # C5, C6

        is_motorway_mask = df["road_type"].isin(["motorway", "trunk"])

        for curve in other_curves:
            df.loc[is_motorway_mask, curve] = np.nan

        for curve in motorway_curves:
            df.loc[~is_motorway_mask, curve] = np.nan

        # Add the new columns add the right location to the df
        self.gdf[all_dam_cols] = df[all_dam_cols]
        logging.info(
            "calculate_damage_OSdaMage(): Damage calculation with the OSdaMage functions was succesfull"
        )

    ### Utils handlers
    def create_mask(self):
        """
        #Create a mask of only the dataframes with hazard data (to speed-up damage calculations)
        effect: *self._gdf_mask* = mask of only the rows with hazard data

        """

        # because the fractions are often 0 (also if the rest is nan, this messes up the .isna)
        val_cols_temp = [c for c in self.val_cols if "_fr" not in c]
        gdf_mask = self.gdf.loc[~(self.gdf[val_cols_temp].isna()).all(axis=1)]
        self._gdf_mask = gdf_mask  # todo: not sure if we need to store the mask

        # Also remove the geometries from the mask
        column_names = list(self._gdf_mask.columns)
        if "geometry" in column_names:
            column_names.remove("geometry")
        self._gdf_mask = self._gdf_mask[column_names]

    def replace_none_with_nan(self):
        import numpy as np

        dam_cols = [c for c in self.gdf.columns if c.startswith("dam_")]
        self.gdf[dam_cols] = self.gdf[dam_cols].fillna(value=np.nan)



"""
                    GNU GENERAL PUBLIC LICENSE
                      Version 3, 29 June 2007

    Risk Assessment and Adaptation for Critical Infrastructure (RA2CE).
    Copyright (C) 2023 Stichting Deltares

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

from __future__ import annotations

from pathlib import Path

from ra2ce.common.configuration.config_wrapper_protocol import ConfigWrapperProtocol
from ra2ce.network.graph_files.graph_files_collection import GraphFilesCollection
from ra2ce.network.hazard.hazard_overlay import HazardOverlay
from ra2ce.network.network_config_data.network_config_data import NetworkConfigData
from ra2ce.network.network_config_data.network_config_data_validator import (
    NetworkConfigDataValidator,
)
from ra2ce.network.networks import Network


class NetworkConfigWrapper(ConfigWrapperProtocol):
    ini_file: Path
    config_data: NetworkConfigData
    graph_files: GraphFilesCollection

    def __init__(self) -> None:
        self.ini_file = None
        self.config_data = NetworkConfigData()
        self.graph_files = GraphFilesCollection()

    @classmethod
    def from_data(
        cls, ini_file: Path, config_data: NetworkConfigData
    ) -> NetworkConfigWrapper:
        """
        Initializes a `NetworkConfig` with the given parameters.

        Args:
            ini_file (Path): Path to the ini file containing the analysis data.
            config_data (NetworkIniConfigData): Ini data representation.

        Returns:
            NetworkConfig: Initialized instance.
        """
        _new_network_config = cls()
        _new_network_config.ini_file = ini_file
        _new_network_config.config_data = config_data
        if config_data.output_graph_dir:
            if config_data.output_graph_dir.is_dir():
                _new_network_config.graph_files = (
                    _new_network_config.read_graphs_from_config(
                        config_data.output_graph_dir
                    )
                )
            else:
                config_data.output_graph_dir.mkdir(parents=True)
        return _new_network_config

    @staticmethod
    def read_graphs_from_config(static_output_dir: Path) -> GraphFilesCollection:
        if not static_output_dir.exists():
            raise ValueError("Path does not exist: {}".format(static_output_dir))
        # Load graphs
        _graph_files = GraphFilesCollection.set_files(static_output_dir)

        return _graph_files

    def configure(self) -> None:
        self.configure_network()
        self.configure_hazard()

    def configure_network(self) -> None:
        network = Network(self.config_data, self.graph_files)
        self.graph_files = network.create()

    def configure_hazard(self) -> None:
        # Call Hazard Handler (to rework)
        if not self.graph_files.has_graphs():
            self.graph_files = self.read_graphs_from_config(
                self.config_data.static_path.joinpath("output_graph")
            )

        if not self.config_data.hazard.hazard_map:
            return

        # There is a hazard map or multiple hazard maps that should be intersected with the graph.
        hazard = HazardOverlay(self.config_data, self.graph_files)
        self.graph_files = hazard.create()

    def is_valid(self) -> bool:
        return NetworkConfigDataValidator(self.config_data).validate().is_valid()



"""
                    GNU GENERAL PUBLIC LICENSE
                      Version 3, 29 June 2007

    Risk Assessment and Adaptation for Critical Infrastructure (RA2CE).
    Copyright (C) 2023 Stichting Deltares

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

from __future__ import annotations

from dataclasses import dataclass, field
from pathlib import Path


@dataclass
class HazardFiles:
    tif: list[Path] = field(default_factory=list)
    gpkg: list[Path] = field(default_factory=list)
    table: list[Path] = field(default_factory=list)

    @classmethod
    def from_hazard_map(cls, hazard_map: list[Path]) -> HazardFiles:
        """
        Create a HazardFiles object from a list of hazard map files.

        Args:
            hazard_map (list[Path]): _description_

        Returns:
            HazardFiles: _description_
        """

        def _get_filtered_files(*suffix) -> list[Path]:
            _filter = lambda x: x.suffix in list(suffix)
            return list(filter(_filter, hazard_map))

        return cls(
            tif=_get_filtered_files(".tif"),
            gpkg=_get_filtered_files(".gpkg"),
            table=_get_filtered_files(".csv"),
        )



"""
                    GNU GENERAL PUBLIC LICENSE
                      Version 3, 29 June 2007

    Risk Assessment and Adaptation for Critical Infrastructure (RA2CE).
    Copyright (C) 2023 Stichting Deltares

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

import logging

from ra2ce.configuration.config_wrapper import ConfigWrapper
from ra2ce.runners.analysis_runner_protocol import AnalysisRunner
from ra2ce.runners.damages_analysis_runner import DamagesAnalysisRunner
from ra2ce.runners.losses_analysis_runner import LossesAnalysisRunner


class AnalysisRunnerFactory:
    @staticmethod
    def get_runner(ra2ce_input: ConfigWrapper) -> AnalysisRunner:
        """
        Gets the supported first analysis runner for the given input. The runner is initialized within this method.

        Args:
            ra2ce_input (Ra2ceInput): Input representing a set of network and analysis ini configurations.

        Returns:
            AnalysisRunner: Initialized Ra2ce analysis runner.
        """
        _available_runners = [DamagesAnalysisRunner, LossesAnalysisRunner]
        _supported_runners = [
            _runner for _runner in _available_runners if _runner.can_run(ra2ce_input)
        ]
        if not _supported_runners:
            _err_mssg = "No analysis runner found for the given configuration."
            logging.error(_err_mssg)
            raise ValueError(_err_mssg)

        # Initialized selected supported runner (First one available).
        _selected_runner = _supported_runners[0]()
        if len(_supported_runners) > 1:
            logging.warning(f"More than one runner available, using {_selected_runner}")
        return _selected_runner



import logging
import os
from pathlib import Path

import geopandas as gpd
import numpy as np
import pandas as pd

from ra2ce.analysis.analysis_config_data.analysis_config_data import (
    AnalysisSectionDamages,
)
from ra2ce.analysis.analysis_input_wrapper import AnalysisInputWrapper
from ra2ce.analysis.damages.analysis_damages_protocol import AnalysisDamagesProtocol
from ra2ce.network.graph_files.network_file import NetworkFile


class EffectivenessMeasures(AnalysisDamagesProtocol):
    analysis: AnalysisSectionDamages
    graph_file: NetworkFile
    graph_file_hazard: NetworkFile
    input_path: Path
    output_path: Path

    def __init__(
        self,
        analysis_input: AnalysisInputWrapper,
    ) -> None:
        self.analysis = analysis_input.analysis
        self.graph_file = None
        self.graph_file_hazard = analysis_input.graph_file_hazard
        self.input_path = analysis_input.input_path
        self.output_path = analysis_input.output_path
        self.result = None

        self.return_period = self.analysis.return_period  # years
        self.repair_costs = self.analysis.repair_costs  # euro
        self.evaluation_period = self.analysis.evaluation_period  # years
        self.interest_rate = self.analysis.interest_rate / 100  # interest rate
        self.btw = 1.21  # VAT multiplication factor to include taxes
        self._measures_csv = self.input_path.joinpath(
            "damages", "effectiveness_measures.csv"
        )

        # perform checks on input while initializing class
        self._validate_input_params(self.input_path, self.analysis)

    def _validate_input_params(
        self, input_path: Path, analysis: AnalysisSectionDamages
    ) -> None:
        if analysis.file_name is None:
            _error = "Effectiveness of measures calculation: No input file configured. Please define an input file in the analysis.ini file."
            logging.error(_error)
            raise ValueError(_error)
        elif analysis.file_name.suffix != ".shp":
            _error = "Effectiveness of measures calculation: Wrong input file configured. Extension of input file is -{}-, needs to be -.shp- (shapefile)".format(
                analysis.file_name.suffix
            )
            logging.error(_error)
            raise ValueError(_error)
        elif not (input_path / "damages" / analysis.file_name).exists():
            _error = "Effectiveness of measures calculation: Input file doesn't exist please place file in the following folder: {}".format(
                input_path / "damages"
            )
            logging.error(_error)
            raise FileNotFoundError(input_path / "damages" / analysis.file_name)
        elif not self._measures_csv.exists():
            _error = "Effectiveness of measures calculation: lookup table with effectiveness of measures doesnt exist. Please place the effectiveness_measures.csv file in the following folder: {}".format(
                input_path / "damages"
            )
            logging.error(_error)
            raise FileNotFoundError(self._measures_csv)

    def _load_effectiveness_table(self) -> dict:
        """This function loads a CSV table containing effectiveness of the different aspects for a number of strategies"""
        df_lookup = pd.read_csv(self._measures_csv, index_col="strategies")
        return df_lookup.transpose().to_dict()

    def _create_feature_table(self, file_path: Path) -> pd.DataFrame:
        """This function loads a table of features from the input folder"""
        logging.info("Loading feature dataframe...")
        gdf = gpd.read_file(file_path, engine="pyogrio")
        logging.info("Dataframe loaded...")

        # cleaning up dataframe
        df = pd.DataFrame(gdf.drop(columns="geometry"))
        df = df[df["LinkNr"] != 0]
        df = df.sort_values(by=["LinkNr", "TARGET_FID"])
        df = df.rename(
            columns={
                "ver_hoog_m": "ver_hoger_m",
                "hwaafwho_m": "hwa_afw_ho_m",
                "slope_15_m": "slope_0015_m",
                "slope_1_m": "slope_001_m",
                "TARGET_FID": "target_fid",
                "Length": "length",
            }
        )
        df = df[
            [
                "LinkNr",
                "target_fid",
                "length",
                "dichtbij_m",
                "ver_hoger_m",
                "hwa_afw_ho_m",
                "gw_hwa_m",
                "slope_0015_m",
                "slope_001_m",
            ]
        ]

        # save as csv
        path, file = os.path.split(file_path)
        df.to_csv(os.path.join(path, file.replace(".gpkg", ".csv")), index=False)
        return df

    def _load_table(self, path: Path, file: str) -> pd.DataFrame:
        """This method reads the dataframe created from"""
        file_path = path / file
        df = pd.read_csv(file_path)
        return df

    def _knmi_correction(self, df: pd.DataFrame, duration: int = 60) -> pd.DataFrame:
        """This function corrects the length of each segment depending on a KNMI factor.
        This factor is calculated using an exponential relation and was calculated using an analysis on all line elements
        a relation is establisched for a 10 minute or 60 minute rainfall period
        With a boolean you can decide to export length or the coefficient itself
        max 0.26 en 0.17
        """
        if duration not in [10, 60]:
            _error_mssg = "Wrong duration configured, has to be 10 or 60"
            logging.error(_error_mssg)
            raise ValueError(_error_mssg)
        logging.info(
            "Applying knmi length correction with duration of rainfall of -{}- minutes".format(
                duration
            )
        )

        coefficients_lookup = {
            10: {"a": 1.004826523, "b": -0.000220199, "max": 0.17},
            60: {"a": 1.012786829, "b": -0.000169182, "max": 0.26},
        }

        coefficient = coefficients_lookup[duration]
        df["coefficient"] = coefficient["a"] * np.exp(coefficient["b"] * df["length"])
        df["coefficient"] = df["coefficient"].where(
            df["length"].astype(float) <= 8000, other=coefficient["max"]
        )
        return df

    def _calculate_effectiveness(
        self, df: pd.DataFrame, name: str = "standard"
    ) -> pd.DataFrame:
        """This function calculates effectiveness, based on a number of columns:
        'dichtbij_m', 'ver_hoger_m', 'hwa_afw_ho_m', 'gw_hwa_m', slope_0015_m' and 'slope_001_m'
        and contains the following steps:
        1. calculate the max of ver_hoger, hwa_afw_ho and gw_hwa columns --> verweg
        2. calculate maximum of slope0015 / 2 and slope 001 columns --> verkant
        3. calculate max of verweg, verkant and dichtbij
        4. calculate sum of verweg, verkant and dichtbij
        5. aggregate (sum) of values to LinkNr
        """
        _gevoelig_max_name = "{}_gevoelig_max".format(name)
        _gevoelig_sum_name = "{}_gevoelig_sum".format(name)

        # perform calculation of max length of ver weg elements and slope elements:
        df["slope_0015_m2"] = df["slope_0015_m"] / 2
        df["verweg_max"] = (
            df[["ver_hoger_m", "hwa_afw_ho_m", "gw_hwa_m"]].values.max(1).round(0)
        )
        df["verkant_max"] = df[["slope_0015_m2", "slope_001_m"]].values.max(1).round(0)

        # calculate gevoelig max and dum
        df[_gevoelig_max_name] = (
            df[["verweg_max", "verkant_max", "dichtbij_m"]].values.max(1).round(0)
        )
        df[_gevoelig_sum_name] = df["verweg_max"] + df["verkant_max"] + df["dichtbij_m"]

        # aggregate to link nr
        new_df = df[
            [
                "LinkNr",
                "length",
                "dichtbij_m",
                "ver_hoger_m",
                "hwa_afw_ho_m",
                "gw_hwa_m",
                "verweg_max",
                "verkant_max",
                _gevoelig_max_name,
                _gevoelig_sum_name,
            ]
        ]
        new_df = new_df.groupby(["LinkNr"]).sum()
        new_df["LinkNr"] = new_df.index
        new_df = new_df.reset_index(drop=True)

        return new_df[
            [
                "LinkNr",
                "length",
                "dichtbij_m",
                "ver_hoger_m",
                "hwa_afw_ho_m",
                "gw_hwa_m",
                "verweg_max",
                "verkant_max",
                _gevoelig_max_name,
                _gevoelig_sum_name,
            ]
        ]

    def _calculate_strategy_effectiveness(
        self, df: pd.DataFrame, effectiveness_dict: dict
    ) -> pd.DataFrame:
        """This function calculates the efficacy for each strategy"""

        columns = [
            "dichtbij",
            "ver_hoger",
            "hwa_afw_ho",
            "gw_hwa",
            "slope_0015",
            "slope_001",
        ]

        # calculate standard effectiveness without factors
        df_total = self._calculate_effectiveness(df, name="standard")

        df_blockage = pd.read_csv(self.input_path / "damages" / "blockage_costs.csv")
        df_total = df_total.merge(df_blockage, how="left", on="LinkNr")
        df_total["length"] = df_total[
            "afstand"
        ]  # TODO Remove this line as this is probably incorrect, just as a check

        # start iterating over different strategies in lookup dictionary
        for strategy in effectiveness_dict:
            logging.info("Calculating effectiveness of strategy: {}".format(strategy))
            lookup_dict = effectiveness_dict[strategy]
            df_temp = df.copy()

            # apply the effectiveness factor as read from the lookup table on each column:
            for col in columns:
                df_temp[col + "_m"] = df_temp[col + "_m"] * (1 - lookup_dict[col])

            # calculate the effectiveness and add as a new column to total dataframe
            df_new = self._calculate_effectiveness(df_temp, name=strategy)
            df_new = df_new.drop(
                columns={
                    "length",
                    "dichtbij_m",
                    "ver_hoger_m",
                    "hwa_afw_ho_m",
                    "gw_hwa_m",
                    "verweg_max",
                    "verkant_max",
                }
            )
            df_total = df_total.merge(df_new, how="left", on="LinkNr")

        return df_total

    def _calculate_cost_reduction(
        self, df: pd.DataFrame, effectiveness_dict: dict
    ) -> pd.DataFrame:
        """This function calculates the yearly costs and possible reduction"""

        strategies = [strategy for strategy in effectiveness_dict]
        strategies.insert(0, "standard")

        # calculate costs
        for strategy in strategies:
            if strategy != "standard":
                df["max_effectiveness_{}".format(strategy)] = 1 - (
                    df["{}_gevoelig_sum".format(strategy)] / df["standard_gevoelig_sum"]
                )
            df["return_period"] = self.return_period * df["coefficient"]
            df["repair_costs_{}".format(strategy)] = (
                df["{}_gevoelig_max".format(strategy)] * self.repair_costs
            )
            # Keys definition to avoid code duplication.
            _blockage_costs_strategy = "blockage_costs_{}".format(strategy)
            _yearly_blockage_costs_strategy = "yearly_blockage_costs_{}".format(
                strategy
            )
            _yearly_repair_costs_strategy = "yearly_repair_costs_{}".format(strategy)
            _total_costs_strategy = "total_costs_{}".format(strategy)

            df[_blockage_costs_strategy] = df["blockage_costs"]
            df[_yearly_repair_costs_strategy] = (
                df["repair_costs_{}".format(strategy)] / df["return_period"]
            )
            if strategy == "standard":
                df[_yearly_blockage_costs_strategy] = (
                    df[_blockage_costs_strategy] / df["return_period"]
                )
            else:
                df[_yearly_blockage_costs_strategy] = (
                    df[_blockage_costs_strategy]
                    / df["return_period"]
                    * (1 - df["max_effectiveness_{}".format(strategy)])
                )
            df[_total_costs_strategy] = (
                df[_yearly_repair_costs_strategy] + df[_yearly_blockage_costs_strategy]
            )
            if strategy != "standard":
                df["reduction_repair_costs_{}".format(strategy)] = (
                    df["yearly_repair_costs_standard"]
                    - df[_yearly_repair_costs_strategy]
                )
                df["reduction_blockage_costs_{}".format(strategy)] = (
                    df["yearly_blockage_costs_standard"]
                    - df[_yearly_blockage_costs_strategy]
                )
                df["reduction_costs_{}".format(strategy)] = (
                    df["total_costs_standard"] - df[_total_costs_strategy]
                )
                df["effectiveness_{}".format(strategy)] = 1 - (
                    df[_total_costs_strategy] / df["total_costs_standard"]
                )
        return df

    def _cost_benefit_analysis(self, effectiveness_dict) -> tuple[pd.DataFrame, dict]:
        """This method performs cost benefit analysis"""

        def _calc_npv(x: pd.DataFrame, cols: list[str]) -> float:
            pv = np.npv(self.interest_rate, [0] + list(x[cols]))
            return pv

        def _calc_npv_factor(factor: float) -> float:
            cols = np.linspace(
                1,
                1 + (factor * self.evaluation_period),
                self.evaluation_period,
                endpoint=False,
            )
            return np.npv(self.interest_rate, [0] + list(cols))

        def _calc_cash_flow(x, cols):
            cash_flow = x[cols].sum() + x["investment"]
            return cash_flow

        df_cba = pd.DataFrame.from_dict(effectiveness_dict).transpose()
        df_cba["strategy"] = df_cba.index
        df_cba = df_cba.drop(
            columns=[
                "dichtbij",
                "ver_hoger",
                "hwa_afw_ho",
                "gw_hwa",
                "slope_0015",
                "slope_001",
            ]
        )
        df_cba["investment"] = df_cba["investment"] * -1

        df_cba["lifespan"] = df_cba["lifespan"].astype(int)
        for col in ["om_pv", "pv", "cash_flow"]:
            df_cba.insert(0, col, 0)

        # add years
        for year in range(1, self.evaluation_period + 1):
            df_cba[str(year)] = df_cba["investment"].where(
                np.mod(year, df_cba["lifespan"]) == 0, other=0
            )
        year_cols = [str(year) for year in range(1, self.evaluation_period + 1)]

        df_cba["om_pv"] = df_cba.apply(lambda x: _calc_npv(x, year_cols), axis=1)
        df_cba["pv"] = df_cba["om_pv"] + df_cba["investment"]
        df_cba["cash_flow"] = df_cba.apply(
            lambda x: _calc_cash_flow(x, year_cols), axis=1
        )
        df_cba["costs"] = df_cba["pv"] * self.btw
        df_cba["costs_pmt"] = (
            np.pmt(
                self.interest_rate, df_cba["lifespan"], df_cba["investment"], when="end"
            )
            * self.btw
        )
        df_cba = df_cba.round(2)

        costs_dict = df_cba[["costs", "on_column"]].to_dict()
        costs_dict["npv_factor"] = _calc_npv_factor(self.climate_factor)

        return df_cba, costs_dict

    def _calculate_strategy_costs(
        self, df: pd.DataFrame, costs_dict: dict
    ) -> pd.DataFrame:
        """Method to calculate costs, benefits with net present value"""

        costs = costs_dict["costs"]
        columns = costs_dict["on_column"]

        def _columns_check(df: pd.DataFrame, columns: list[str]) -> bool:
            cols_check = []
            for col in columns:
                cols_check.extend(columns[col].split(";"))
            df_cols = list(df.columns)

            if any([True for col in cols_check if col not in df_cols]):
                cols = [col for col in cols_check if col not in df_cols]
                _error_mssg = "Wrong column configured in effectiveness_measures csv file. column {} is not available in imported sheet.".format(
                    cols
                )
                logging.error(_error_mssg)
                raise ValueError(_error_mssg)
            return True

        _columns_check(df, columns)
        strategies = {col: columns[col].split(";") for col in columns}

        for strategy in strategies:
            df["{}_benefits".format(strategy)] = (
                df["reduction_costs_{}".format(strategy)] * costs_dict["npv_factor"]
            )
            select_col = strategies[strategy]
            # Definition of key strings to avoid code duplication
            _strategy_costs = "{}_costs".format(strategy)

            if len(select_col) == 1:
                df[_strategy_costs] = df[select_col[0]] * costs[strategy] * -1 / 1000
            if len(select_col) > 1:
                df[_strategy_costs] = (
                    (df[select_col[0]] - df[select_col[1]])
                    * costs[strategy]
                    * -1
                    / 1000
                )
                df[_strategy_costs] = df[_strategy_costs].where(
                    df[_strategy_costs] > 1, other=np.nan
                )
            df["{}_bc_ratio".format(strategy)] = (
                df["{}_benefits".format(strategy)] / df[_strategy_costs]
            )

        return df

    def execute(self) -> None:
        """This function calculated the efficiency of measures. Input is a csv file with efficiency
        and a list of different aspects you want to check.
        """
        effectiveness_dict = self._load_effectiveness_table()

        gdf_in = self.graph_file_hazard.get_graph()

        if self.analysis.create_table:
            df = self._create_feature_table(
                self.input_path.joinpath("damages", self.analysis.file_name)
            )
        else:
            df = self._load_table(
                self.input_path.joinpath("damages"),
                self.analysis.file_name.replace(".gpkg", ".csv"),
            )

        df = self._calculate_strategy_effectiveness(df, effectiveness_dict)
        df = self._knmi_correction(df)
        df_cba, costs_dict = em.cost_benefit_analysis(effectiveness_dict)
        df_cba.round(2).to_csv(
            self.output_path.joinpath(
                self.analysis.analysis.config_value,
                "cost_benefit_analysis.csv",
            ),
            decimal=",",
            sep=";",
            index=False,
            float_format="%.2f",
        )
        df = self._calculate_cost_reduction(df, effectiveness_dict)
        df_costs = self._calculate_strategy_costs(df, costs_dict)
        df_costs = df_costs.astype(float).round(2)
        df_costs.to_csv(
            self.output_path.joinpath(
                self.analysis.analysis.config_value, "output_analysis.csv"
            ),
            decimal=",",
            sep=";",
            index=False,
            float_format="%.2f",
        )

        return gdf_in.merge(df, how="left", on="LinkNr")



from ra2ce.configuration.ra2ce_enum_base import Ra2ceEnumBase


class DamageCurveEnum(Ra2ceEnumBase):
    HZ = 1
    OSD = 2
    MAN = 3
    INVALID = 99

    @property
    def config_value(self) -> str | None:
        """
        Reconstruct the name as it is known in the config.
        This could entail replacement of " " by "_" and lower() operations.

        Returns:
            str: Value as known in the config.
        """
        if self.name == "NONE":
            return None
        return self.name



"""
                    GNU GENERAL PUBLIC LICENSE
                      Version 3, 29 June 2007
    Risk Assessment and Adaptation for Critical Infrastructure (RA2CE).
    Copyright (C) 2023 Stichting Deltares
    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""
from ast import literal_eval
from pathlib import Path
from re import findall

import pandas as pd

from ra2ce.analysis.losses.losses_input_data_reader_base import (
    LossesInputDataReaderBase,
)
from ra2ce.analysis.losses.resilience_curves.resilience_curves import ResilienceCurves
from ra2ce.network.network_config_data.enums.road_type_enum import RoadTypeEnum


class ResilienceCurvesReader(LossesInputDataReaderBase):
    """
    Class to read the resilience curves from a csv file.
    """

    csv_columns = [
        "link_type_hazard_intensity",
        "duration_steps",
        "functionality_loss_ratio",
    ]
    data_type = ResilienceCurves

    def _parse_df(self, df: pd.DataFrame) -> ResilienceCurves:
        def parse_link_type_hazard_intensity(
            link_type_hazard_intensity: str,
        ) -> tuple[str, str, str]:
            return findall(r"^(\w+)_([\d.]+)-([\d.]+)$", link_type_hazard_intensity)[0]

        _resilience_curves = {}
        for _, row in df.iterrows():
            _link_type_hazard_intensity = row["link_type_hazard_intensity"]
            _lt, _h_min, _h_max = parse_link_type_hazard_intensity(
                _link_type_hazard_intensity
            )
            _ds_list = literal_eval(row["duration_steps"])
            _flr_list = literal_eval(row["functionality_loss_ratio"])
            if len(_ds_list) != len(_flr_list):
                raise ValueError(
                    f"Duration steps and functionality loss ratio should have the same length ({_link_type_hazard_intensity})."
                )
            _resilience_curves[
                (RoadTypeEnum.get_enum(_lt), (float(_h_min), float(_h_max)))
            ] = [(float(_ds), float(_flr)) for _ds, _flr in zip(_ds_list, _flr_list)]

        return ResilienceCurves(resilience_curves=_resilience_curves)

    def read(self, file_path: Path | None) -> ResilienceCurves:
        return super().read(file_path)



"""
                    GNU GENERAL PUBLIC LICENSE
                      Version 3, 29 June 2007

    Risk Assessment and Adaptation for Critical Infrastructure (RA2CE).
    Copyright (C) 2023 Stichting Deltares

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

import logging
from pathlib import Path
from typing import Optional

import geopandas as gpd
import networkx as nx
import numpy as np
from geopandas import GeoDataFrame
from osmnx import graph_to_gdfs
from shapely.geometry import LineString, Point

from ra2ce.network import networks_utils as nut


def from_shapefile_to_poly(shapefile: Path, out_path: Path, outname: str = ""):
    """
    This function will create the .poly files from an input shapefile.
    If the shapefile contains multiple polygons, this function creates a seperate .polygon file for each region
    .poly files can then be used to extract data from the openstreetmap files.

    This function is adapted from the OSMPoly function in QGIS, and Elco Koks GMTRA model.
    This code is maintained on a GitHub repository: github.com/keesvanginkel/OSdaMage

    Arguments:
        *shapefile* (string/Pathlib Path) : path to the shapefile
        *out_path* (string/Pathlib Path): path to the directory where the .poly files should be written
        *outname* (string) : optional prefix to add to outfile name

    Returns:
        .poly file for each region, in a new dir in the working directory (in the CRS of te input file)
    """
    shp_file_gdf = gpd.read_file(str(shapefile), engine="pyogrio")

    num = 0
    # iterate over the seperate polygons in the shapefile
    for f in shp_file_gdf.iterrows():
        f = f[1]
        num = num + 1
        geom = f.geometry

        try:
            # this will create a list of the different subpolygons
            if geom.geom_type == "MultiPolygon":
                polygons = geom

            # the list will be length 1 if it is just one polygon
            elif geom.geom_type == "Polygon":
                polygons = [geom]

            # define the name of the output file
            id_name = str(f.name)

            # start writing the .poly file
            _poly_filename = out_path / (outname + id_name + ".poly")
            with open(_poly_filename, "w") as _poly_file:
                _poly_file.write(id_name + "\n")
                i = 0

                # loop over the different polygons, get their exterior and write the
                # coordinates of the ring to the .poly file
                for polygon in polygons:
                    polygon = np.array(polygon.exterior)

                    j = 0
                    _poly_file.write(str(i) + "\n")

                    for ring in polygon:
                        j = j + 1
                        _poly_file.write(
                            "    " + str(ring[0]) + "     " + str(ring[1]) + "\n"
                        )

                    i = i + 1
                    # close the ring of one subpolygon if done
                    _poly_file.write("END" + "\n")

                # close the file when done
                _poly_file.write("END" + "\n")
        except Exception as e:
            print("Exception {}".format(e))


def graph_to_gdf(
    graph: nx.Graph,
    nodes: bool,
    edges: bool,
    node_geometry: bool,
    fill_edge_geometry: bool,
) -> GeoDataFrame:
    u, v, k, data = zip(*graph.edges(keys=True, data=True))
    graph_gdf = graph_to_gdfs(graph, nodes, edges, node_geometry, fill_edge_geometry)
    graph_gdf["data"] = data
    return graph_gdf


def get_node_nearest_edge(
    graph: nx.MultiDiGraph, node: tuple, return_geom=True, return_dist=True
) -> dict:
    """
    Based on osmnx.
    """
    node_coor = (node[1]["x"], node[1]["y"])
    # get u, v, key, geom from all the graph edges
    gdf_edges = graph_to_gdf(
        graph, nodes=False, edges=True, node_geometry=True, fill_edge_geometry=True
    )
    if {"u", "v", "key"}.issubset(gdf_edges.index.names):
        gdf_edges.reset_index(inplace=True)
    edges = gdf_edges[["u", "v", "data", "key", "geometry"]].values

    # convert lat,lng (y,x) node to x,y for shapely distance operation
    xy_point = Point(node_coor)

    # calculate Euclidean distance from each edge's geometry to this node
    edge_distances = [(edge, xy_point.distance(edge[4])) for edge in edges]

    # the nearest edge minimizes the non-zero distance to the node
    (u, v, data, key, geom), dist = min(
        edge_distances, key=lambda x: x[1] if x[1] > 0 else float("inf")
    )
    logging.info(f"Found nearest edge ({u, v, data, key}) to node {node}")

    # return results requested by caller
    if data == 0 and key == 0:
        if return_dist and return_geom:
            return {"node": node, "nearest_edge": (u, v, geom, round(dist, 7))}
        elif return_dist:
            return {"node": node, "nearest_edge": (u, v, round(dist, 7))}
        elif return_geom:
            return {"node": node, "nearest_edge": (u, v, geom)}
        else:
            return {"node": node, "nearest_edge": (u, v)}
    elif data == 0:
        if return_dist and return_geom:
            return {"node": node, "nearest_edge": (u, v, key, geom, round(dist, 7))}
        elif return_dist:
            return {"node": node, "nearest_edge": (u, v, key, round(dist, 7))}
        elif return_geom:
            return {"node": node, "nearest_edge": (u, v, key, geom)}
        else:
            return {"node": node, "nearest_edge": (u, v, key)}
    elif key == 0:
        if return_dist and return_geom:
            return {"node": node, "nearest_edge": (u, v, data, geom, round(dist, 7))}
        elif return_dist:
            return {"node": node, "nearest_edge": (u, v, data, round(dist, 7))}
        elif return_geom:
            return {"node": node, "nearest_edge": (u, v, data, geom)}
        else:
            return {"node": node, "nearest_edge": (u, v, data)}
    else:
        if return_dist and return_geom:
            return {
                "node": node,
                "nearest_edge": (u, v, data, key, geom, round(dist, 7)),
            }
        elif return_dist:
            return {"node": node, "nearest_edge": (u, v, data, key, round(dist, 7))}
        elif return_geom:
            return {"node": node, "nearest_edge": (u, v, data, key, geom)}
        else:
            return {"node": node, "nearest_edge": (u, v, data, key)}


def is_endnode_check(graph: nx.MultiDiGraph, node_id: int) -> bool:
    """
    Based on osmnx. osmnx rules 3 and 4 are removed. Hence, the name is_endpoint_simplified.
    Determine if a node is a true endpoint of an edge.

    Return True if the node is a "true" endpoint of an edge in the network,
    otherwise False. OpenStreetMap's data includes many nodes that exist only as
    geometric vertices to allow ways to curve. A true edge endpoint is a node
    that satisfies at least 1 of the following 4 rules:

    1) It is its own neighbor (ie, it self-loops).

    2) Or, it has no incoming edges or no outgoing edges (ie, all its incident
    edges are inbound or all its incident edges are outbound).

    graph : networkx.MultiDiGraph input graph
    node_id : int the node to examine
    """
    neighbors = set(list(graph.predecessors(node_id)) + list(graph.successors(node_id)))
    # rule 1
    if node_id in neighbors:
        # if the node appears in its list of neighbors, it self-loops
        # this is always an endpoint.
        return True

    # rule 2
    elif graph.out_degree(node_id) == 0 or graph.in_degree(node_id) == 0:
        # if node has no incoming edges or no outgoing edges, it is an endpoint
        return True
    else:
        return False


def create_edges(
    graph: nx.MultiDiGraph, u: int, v: int, key_data: dict
) -> nx.MultiDiGraph:
    """
    Create or update edges between nodes `u` and `v` in the graph with the given data.

    Args:
        graph (nx.MultiDiGraph): The graph to modify.
        u (int): The ID of the source node.
        v (int): The ID of the target node.
        data (dict): The attributes for the new edge(s).

    Returns:
        nx.MultiDiGraph: The modified graph.
    """
    # Ensure all keys in data are strings
    for key, data in key_data.items():
        if not graph.has_edge(u, v, key):
            geom_data = {
                "geometry": LineString(
                    [
                        graph.nodes(data=True)[u]["geometry"],
                        graph.nodes(data=True)[v]["geometry"],
                    ]
                )
            }
            graph.add_edge(u, v, **data)
            graph[u][v][key].update(geom_data)
    return graph


def update_edges_by_node(
    graph: nx.MultiDiGraph, node_edges: list, node_id: int, new_node_geom: dict
) -> None:
    """
    Update connected edges of a node by replacing node_id and node_geom with new_node_id and new_node_geom.

    Args:
        node_edges: edges connected to the node in teh original graph to modify
        graph (nx.MultiDiGraph): The graph to modify.
        node_id (int): The ID of the node to update.
        new_node_geom (dict): Dictionary containing geometry information for the new node.

    Returns:
        None
    """
    # Iterate over connected edges
    for u, v, key, existing_edge_data in node_edges:
        # Step 0: Save existing edge data (excluding geometry and length)
        new_edge_data = existing_edge_data.copy()
        # Step 1: Determine if node_id is a u or v in the edge
        if u == node_id:
            new_edge_id = (node_id, v, key)
            new_line_points = LineString(
                [new_node_geom, graph.nodes(data=True)[v]["geometry"]]
            )
        elif v == node_id:
            new_edge_id = (u, node_id, key)
            new_line_points = LineString(
                [graph.nodes(data=True)[u]["geometry"]], new_node_geom
            )

        # Step 2: Set attributes of the new edge using existing edge data and
        # Set geometry and length of the new edge
        new_edge_data.update(
            {
                "geometry": LineString(new_line_points),
                "length": nut.line_length(
                    LineString(new_line_points), graph.graph["crs"]
                ),
            }
        )

        # Step 3: Create a new edge by replacing node_id with new_node_id
        graph.add_edge(*new_edge_id, **new_edge_data)


def modify_graph(graph: nx.MultiDiGraph, node_nearest_edge_data: dict):
    def set_edge_length(
        graph: nx.MultiDiGraph, u_nearest_edge: int, v_nearest_edge, new_node_id
    ):
        for _, nodes in enumerate(
            [(u_nearest_edge, new_node_id), (new_node_id, v_nearest_edge)], start=1
        ):
            u, v = nodes
            for key in graph[u][v]:
                length_data = {
                    "length": nut.line_length(
                        graph[u][v][key]["geometry"], graph.graph["crs"]
                    )
                }
                graph[u][v][key].update(length_data)

    def get_edges_by_node(graph, node):
        return list(graph.edges(node, data=True, keys=True))

    # step 1: set variables
    u_nearest_edge = node_nearest_edge_data["nearest_edge"][0]
    v_nearest_edge = node_nearest_edge_data["nearest_edge"][1]
    nearest_edge_geom = node_nearest_edge_data["nearest_edge"][3]
    nearest_edge_key_data = graph.get_edge_data(u_nearest_edge, v_nearest_edge).copy()
    if nearest_edge_key_data is None:
        raise ValueError("Edge not found in the graph")

    node_id = node_nearest_edge_data["node"][0]
    node_geom = node_nearest_edge_data["node"][1]["geometry"]
    node_edges = get_edges_by_node(graph, node_id).copy()

    new_node_geom = nearest_edge_geom.interpolate(nearest_edge_geom.project(node_geom))
    new_node_data = node_nearest_edge_data["node"][1].copy()
    new_node_data = remove_key(new_node_data, ["geometry", "x", "y"])

    # step 2: add the new_node
    # Check if the new_node already exists in the nodes of the graph
    if find_existing_node(graph, new_node_geom)[0] is None:
        graph.remove_node(node_id)
        graph.add_node(
            node_id,
            x=new_node_geom.x,
            y=new_node_geom.y,
            geometry=new_node_geom,
            **new_node_data,
        )
    else:
        node_id = find_existing_node(graph, new_node_geom)[0]

    # step 3: update edges of the graph being affected by adding a new node
    graph.remove_edge(u_nearest_edge, v_nearest_edge)
    # Add new node to the nearest edge and create two edges
    graph = create_edges(graph, u_nearest_edge, node_id, nearest_edge_key_data)
    graph = create_edges(graph, node_id, v_nearest_edge, nearest_edge_key_data)
    # update the length of the created edges on the nearest_edge
    set_edge_length(graph, u_nearest_edge, v_nearest_edge, node_id)

    # update edges connected to the node close enough (passed the threshold) to the nearest edge
    update_edges_by_node(graph, node_edges, node_id, new_node_geom)


def remove_key(element_data: dict, keys_to_exclude: list) -> dict:
    """
    Removes keys such as geometry information from the new node data to be created
    Args:
        element_data: dict, node or edge data of a nx.Graph object
        keys_to_exclude: list, information key to exclude from the element_data

    Returns: dict, filtered element_data for the new node or edge to be created in a new nx.Graph object

    """

    for key in keys_to_exclude:
        element_data.pop(key)
    return element_data


def find_existing_node(
    graph: nx.MultiDiGraph, new_node: Point
) -> tuple[int, dict] | tuple[None, None]:
    """
    finds whether a newly created node exists in a graph
    Args:
        graph: nx.MultiDiGraph
        new_node: Shapely Point, a newly created node

    Returns: None if a newly created node does not exist in a graph. Otherwise, returns the node and its data

    """
    for node, data in graph.nodes(data=True):
        if not data.get("geometry", ""):
            raise ValueError("Nodes should have geometry")

        if data["geometry"] == new_node:
            return node, data
    return None, None


def create_edge(graph: nx.MultiDiGraph, u: int, v: int, data: dict) -> nx.MultiDiGraph:
    """
    creates an edge if it does not already exist.
    Args:
        graph: nx.MultiDiGraph
        u: int, starting node of an edge
        v: int, ending node of an edge

    Returns: updated/existing graph containing a new edge

    """
    if graph.has_edge(u, v):
        return graph
    else:
        graph.add_edge(u, v, **data[0])
        return graph



"""
                    GNU GENERAL PUBLIC LICENSE
                      Version 3, 29 June 2007

    Risk Assessment and Adaptation for Critical Infrastructure (RA2CE).
    Copyright (C) 2023 Stichting Deltares

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

# -*- coding: utf-8 -*-
import logging
from pathlib import Path
from typing import Optional

import click

from ra2ce.analysis.analysis_result_wrapper import AnalysisResultWrapper
from ra2ce.ra2ce_handler import Ra2ceHandler


### Below is the documentation for the commandline interface, see the CLICK-package.
@click.command()
@click.option("--network_ini", default=None, help="Full path to the network.ini file.")
@click.option(
    "--analyses_ini", default=None, help="Full path to the analyses.ini file."
)
def run_analysis(network_ini: str, analyses_ini: str) -> list[AnalysisResultWrapper]:
    def _as_path(ini_file: str) -> Optional[Path]:
        if not ini_file:
            return None

        _ini = Path(ini_file)
        if not _ini.is_file():
            raise FileNotFoundError(_ini)
        return _ini

    _network_ini = _as_path(network_ini)
    _analysis_ini = _as_path(analyses_ini)
    _handler = Ra2ceHandler(_network_ini, _analysis_ini)
    _handler.configure()
    return _handler.run_analysis()


if __name__ == "__main__":
    try:
        run_analysis()
    except Exception as e_info:
        logging.error(str(e_info))
        raise



"""
                    GNU GENERAL PUBLIC LICENSE
                      Version 3, 29 June 2007
    Risk Assessment and Adaptation for Critical Infrastructure (RA2CE).
    Copyright (C) 2023 Stichting Deltares
    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

import logging
from dataclasses import dataclass

import networkx as nx
from osmnx.simplification import simplify_graph

from ra2ce.network.networks_utils import add_x_y_to_nodes

NxGraph = nx.Graph | nx.MultiGraph | nx.MultiDiGraph


@dataclass(kw_only=True)
class NetworkSimplificationWithoutAttributeExclusion:
    """
    Simplifies a network with the `osmnx.simplification` functionality.
    """

    nx_graph: NxGraph

    def simplify_graph(
        self,
    ) -> nx.Graph:
        """
        Simplify the graph after adding missing x and y attributes to nodes

        Returns:
            nx.Graph: Simplified graph
        """
        _complex_graph = add_x_y_to_nodes(self.nx_graph)
        _simple_graph = simplify_graph(
            _complex_graph, strict=True, remove_rings=True, track_merged=False
        )

        logging.info(
            "Graph simplified from %s to %s nodes and %s to %s edges.",
            _complex_graph.number_of_nodes(),
            _simple_graph.number_of_nodes(),
            _complex_graph.number_of_edges(),
            _simple_graph.number_of_edges(),
        )

        return _simple_graph



"""
                    GNU GENERAL PUBLIC LICENSE
                      Version 3, 29 June 2007

    Risk Assessment and Adaptation for Critical Infrastructure (RA2CE).
    Copyright (C) 2023 Stichting Deltares

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""
import itertools
import logging
import math
import os
import sys
import warnings
from pathlib import Path
from statistics import mean
from typing import Optional

import geopandas as gpd
import networkx as nx
import numpy as np
import pandas as pd
import pyproj
import rasterio
import rtree
from geopy import distance
from numpy.ma import MaskedArray
from osgeo import gdal
from osmnx import graph_to_gdfs
from rasterio.features import shapes
from rasterio.mask import mask
from shapely.geometry import LineString, MultiLineString, Point, box, shape
from shapely.geometry.base import BaseGeometry, BaseMultipartGeometry
from shapely.ops import linemerge, unary_union
from tqdm import tqdm

from ra2ce.network.network_config_data.enums.road_type_enum import RoadTypeEnum


def convert_unit(unit: str) -> Optional[float]:
    """Converts unit to meters.

    Args:
        unit (str): The unit to convert

    Returns:
        Optional[float]: The result of the conversion.
    """
    _conversion_dict = dict(centimeters=1 / 100, meters=1, feet=1 / 3.28084)
    return _conversion_dict.get(unit.lower(), None)


def draw_progress_bar(percent: float, bar_length: int = 20):
    """Draws a progress bar
    https://stackoverflow.com/questions/3002085/python-to-print-out-status-bar-and-percentage
    """
    # percent float from 0 to 1.
    sys.stdout.write("\r")
    sys.stdout.write(
        "[{:<{}}] {:.0f}%".format(
            "=" * int(bar_length * percent), bar_length, percent * 100
        )
    )
    sys.stdout.flush()


def merge_lines_automatic(
    lines_gdf: gpd.GeoDataFrame, id_name: str, aadt_names: list[str], crs_: pyproj.CRS
) -> tuple[gpd.GeoDataFrame, gpd.GeoDataFrame]:
    """Automatically merge lines based on a config file
    Args:
        lines_gdf (geodataframe): the network with edges that can possibly be merged
        id_name (string): name of the Unique ID column in the lines_gdf
        aadt_names (list of strings): names of the columns of the AADT (average annual daily traffic)
        crs_ (int): the EPSG number of the coordinate reference system that is used
    Returns:
        lines_gdf (geodataframe): the network with edges that are (not) merged
        lines_merged (geodataframe): the lines that are merged, if lines are merged. Otherwise it returns an empty GDF
    """
    list_lines = list(lines_gdf["geometry"])

    # Try to merge the lines
    try:
        merged_lines = linemerge(list_lines)
    except NotImplementedError as e:
        logging.error(
            "Your data contains Multi-part geometries, you cannot merge lines. Error: {}".format(
                e
            )
        )
        return lines_gdf, gpd.GeoDataFrame()
    # merge_input ='y'

    # continue
    if len(merged_lines.geoms) < len(list_lines) and aadt_names:
        # the number of merged lines is smaller than the number of lines in the input, so lines can be merged
        yes_to_all = "all"  # or in ['all', 'single']
        # ask the user if they want to take the max, min or average for all values
        all_type = "max"  # or ['max', 'min', 'mean']

    elif len(merged_lines.geoms) < len(list_lines) and not aadt_names:
        # the_input = 'y'
        lines_gdf = lines_gdf.dissolve(by=id_name, aggfunc="max")
        lines_gdf.reset_index(inplace=True)

    elif len(merged_lines.geoms) == len(list_lines):
        logging.warning("No lines are merged.")
        return lines_gdf, gpd.GeoDataFrame()
    else:
        # The lines have no additional properties.
        return lines_gdf, gpd.GeoDataFrame()

    # Check which of the lines are merged, also for the fid. The fid of the first line with a traffic count is taken.
    # The list of fid's is reduced by the fid's that are not anymore in the merged lines
    lines_fids = list(
        zip(list_lines, lines_gdf[id_name])
    )  # create tuples from the list of lines and the list of fid's
    lines_merged = gpd.GeoDataFrame(
        columns=[id_name, "geometry"], crs=crs_, geometry="geometry"
    )
    merged = gpd.GeoDataFrame(
        columns=[id_name, "geometry"], crs=crs_, geometry="geometry"
    )

    for mline in merged_lines.geoms:
        for line, i in lines_fids:
            full_line = line
            if line.within(mline) and not line.equals(mline):
                # if the line is within the merged line but is not the same, the line is part of a merged line
                line_set_merged = [line]
                fid_set_merged = [i]

                if aadt_names:
                    aadts_set_merged = [
                        lines_gdf[lines_gdf[id_name] == i][aadt_names].iloc[0].tolist()
                    ]
                else:
                    aadts_set_merged = []

                while not full_line.equals(mline):
                    for line2, j in lines_fids:
                        if (
                            line2.within(mline)
                            and not line2.equals(mline)
                            and line != line2
                        ):
                            line_set_merged.append(line2)
                            fid_set_merged.append(j)
                            if aadt_names:
                                aadts_set_merged.append(
                                    lines_gdf[lines_gdf[id_name] == i][aadt_names]
                                    .iloc[0]
                                    .tolist()
                                )

                            lines_fids.remove(
                                (line2, j)
                            )  # remove the lines that have been through the iteration so there are no duplicates
                    full_line = linemerge(line_set_merged)
                lines_fids.remove(
                    (line, i)
                )  # remove the lines that have been through the iteration so there are no duplicates
                # the lines in this list are the same lines that make up the merged line
                add_idx = 0 if lines_merged.empty else max(lines_merged.index) + 1
                lines_merged.loc[add_idx] = {id_name: i, "geometry": full_line}

                # check with the user the right traffic count for the merged lines
                if aadts_set_merged:  # check if the list is not empty
                    aadts_set_merged = [
                        i for i in aadts_set_merged if not all(v is None for v in i)
                    ]
                    if (
                        len(aadts_set_merged) > 1
                        and isinstance(aadts_set_merged[0], list)
                        and yes_to_all == "all"
                    ):
                        if all_type == "max":
                            aadts_set_merged = [
                                max(sublist)
                                for sublist in list(map(list, zip(*aadts_set_merged)))
                            ]
                        elif all_type == "min":
                            aadts_set_merged = [
                                min(sublist)
                                for sublist in list(map(list, zip(*aadts_set_merged)))
                            ]
                        elif all_type == "mean":
                            aadts_set_merged = [
                                mean(sublist)
                                for sublist in list(map(list, zip(*aadts_set_merged)))
                            ]

                # add values to the dataframe
                this_fid = [x[0] if isinstance(x, list) else x for x in fid_set_merged][
                    0
                ]  # take the first feature ID for the merged lines

                # initiate dict for new row in merged gdf
                properties_dict = {id_name: this_fid, "geometry": mline}

                if aadt_names:
                    if isinstance(aadts_set_merged[0], list):
                        this_aadts = aadts_set_merged[0]
                    else:
                        this_aadts = aadts_set_merged

                    # update dict for new row in merged gdf
                    properties_dict.update(
                        {a: aadt_val for a, aadt_val in zip(aadt_names, this_aadts)}
                    )

                # append row to merged gdf
                add_idx = 0 if merged.empty else max(merged.index) + 1
                merged.loc[add_idx] = properties_dict

            elif line.equals(mline):
                # this line is not merged
                properties_dict = {id_name: i, "geometry": mline}

                if aadt_names:
                    properties_dict.update(
                        {
                            a: aadt_val
                            for a, aadt_val in zip(
                                aadt_names,
                                lines_gdf.loc[lines_gdf[id_name] == i][aadt_names].iloc[
                                    0
                                ],
                            )
                        }
                    )
                add_idx = 0 if merged.empty else max(merged.index) + 1
                merged.loc[add_idx] = properties_dict

    merged["length"] = merged["geometry"].apply(lambda x: line_length(x, crs_))

    return merged, lines_merged


def get_distance(a_b_tuple: tuple[tuple[float, float], tuple[float, float]]) -> float:
    """
    Gets the distance (in meters) between two points given as a tuple thanks to geopy.distance functionality.
    TODO: Investigate whether this function is really required, instead of GeoPandasDataFrame,
     to reduce the usage of external dependencies.

    Args:
        a_b_tuple (tuple[float]): Tuple representing two points.

    Returns:
        float: Distance between two points in meters.
    """
    distance.geodesic.ELLIPSOID = "WGS-84"
    from_a, to_b = a_b_tuple
    latlon = lambda lonlat: (lonlat[1], lonlat[0])
    return distance.distance(latlon(from_a), latlon(to_b)).meters


def line_length(line: LineString, crs: pyproj.CRS) -> float:
    """Calculate length of a line in meters, given in geographic coordinates.

    Args:
        line: a shapely LineString object with coordinate reference system 'crs'
        crs: the coordinate reference system of the 'line' LineString

    Returns:
        Length of line in m
    """
    # Check if the coordinate system is projected or geographic
    if crs.is_geographic:
        try:
            # Swap shapely (lonlat) to geopy (latlon) points
            if isinstance(line, LineString):
                total_length = sum(map(get_distance, zip(line.coords, line.coords[1:])))
            elif isinstance(line, MultiLineString):
                total_length = sum(
                    [
                        sum(map(get_distance, zip(l.coords, l.coords[1:])))
                        for l in line.geoms
                    ]
                )
            else:
                logging.error(
                    "The road strech is not a Shapely LineString or MultiLineString so the length cannot be computed."
                    "Please check your data network data."
                )
                return np.nan
        except Exception:
            logging.error(
                "The CRS is not EPSG:4326. Quit the analysis, reproject the layer to EPSG:4326 and try again to run the tool."
            )
            return np.nan
    elif crs.is_projected:
        # line length of projected linestrings
        if isinstance(line, LineString | MultiLineString):
            total_length = line.length
        else:
            logging.error(
                "The road stretch is not a Shapely LineString or MultiLineString so the length cannot be computed."
                "Please check your data network data."
            )
            return np.nan
    return round(total_length, 0)


def snap_endpoints_lines(
    lines_gdf: gpd.GeoDataFrame,
    max_dist: int | float,
    id_name: str,
    crs: pyproj.CRS,
) -> gpd.GeoDataFrame:
    """Snap endpoints of lines with endpoints or vertices of other lines
    if they are at most max_dist apart. Choose the closest endpoint or vertice.

    Args:
        lines_gdf: a list of LineStrings or a MultiLineString
        max_dist: maximum distance two endpoints may be joined together
        id_name: the name of the ID column in lines_gdf

    From shapely_tools:
        @author: Dirk Eilander (dirk.eilander@deltares.nl)
        Adjusted 15-10-2019: Frederique de Groen (frederique.degroen@deltares.nl)
        Build on library from https://github.com/ojdo/python-tools/blob/master/shapelytools.py
    """
    logging.info("Started snapping endpoints of lines...")
    max_id = max(lines_gdf[id_name])

    # initialize snapped lines with list of original lines
    # snapping points is a MultiPoint object of all vertices
    snapped_lines = [line for line in list(lines_gdf["geometry"])]
    snapping_dict = vertices_from_lines(snapped_lines, list(lines_gdf[id_name]))

    # isolated endpoints are being snapped to the closest vertex
    isolated_endpoints = find_isolated_endpoints(
        list(lines_gdf[id_name]), snapped_lines
    )

    logging.info(
        "Number of isolated endpoints (points that probably need to be snapped): {} ".format(
            len(isolated_endpoints)
        )
    )
    logging.info("Snapping lines.. Follow the progress:")
    # only snap isolated endpoints within max_dist of another vertice / endpoint
    for i, isolated_endpoint in enumerate(isolated_endpoints):
        ids, endpoint = isolated_endpoint

        # create a list of the vertices that are not the line's own vertices
        points_without_linepoints = [
            value for key, value in snapping_dict.items() if key != ids
        ]

        # create list of all points to search in
        all_vertices = [p for sublist in points_without_linepoints for p in sublist]

        # create an empty spatial index object to search in
        idx = rtree.index.Index()

        # populate the spatial index
        for j, pnt in enumerate(all_vertices):
            idx.insert(j, pnt.bounds)

        # find all vertices within a radius of max_distance as possible
        # choose closest vertice and line the vertice lays on
        target = nearest_neighbor_within(all_vertices, idx, endpoint, max_dist)

        # draw a progress bar
        draw_progress_bar(i / len(isolated_endpoints))

        # do nothing if the target point is further away from the endpoint than max_dist
        # or if they are at the same location
        if not target:
            continue

        # check if the line does not yet exist
        new_line = LineString([(target.x, target.y), (endpoint.x, endpoint.y)])
        if not any(new_line.equals(another_line) for another_line in snapped_lines):
            if new_line.length > 0:
                lines_gdf = lines_gdf.append(
                    {
                        id_name: max_id + 1,
                        "geometry": new_line,
                        "length": line_length(new_line, crs),
                    },
                    ignore_index=True,
                )
                max_id += 1

    # TODO: remove any lines that are overlapping?

    return lines_gdf


def find_isolated_endpoints(lines_ids: list[str], lines: list) -> list:
    """Find endpoints of lines that don't touch another line.

    Args:
        lines_ids: a list of the IDs of lines
        lines: a list of LineStrings or a MultiLineString

    Returns:
        A list of line end Points that don't touch any other line of lines

    From shapely_tools:
        @author: Dirk Eilander (dirk.eilander@deltares.nl)
        Adjusted 15-10-2019: Frederique de Groen (frederique.degroen@deltares.nl)
        Build on library from https://github.com/ojdo/python-tools/blob/master/shapelytools.py
    """
    isolated_endpoints = []
    for i, id_line in enumerate(zip(lines_ids, lines)):
        ids, line = id_line
        other_lines = lines[:i] + lines[i + 1 :]
        for q in [0, -1]:
            if isinstance(line, LineString):
                endpoint = Point(line.coords[q])
                if any(endpoint.touches(another_line) for another_line in other_lines):
                    continue
                else:
                    isolated_endpoints.append((ids, endpoint))
            elif isinstance(line, MultiLineString):
                endpoints = [Point(l.coords[q]) for l in line]
                for endpnt in endpoints:
                    if any(
                        endpnt.touches(another_line) for another_line in other_lines
                    ):
                        continue
                    else:
                        isolated_endpoints.append((ids, endpnt))
    return isolated_endpoints


def nearest_neighbor_within(
    search_points: list, spatial_index, point: Point, max_distance: float | int
) -> Point:
    """Find nearest point among others up to a maximum distance.

    Args:
        search_points: list of points to search in
        spatial_index: rtree spatial index of the points to search in
        point: a Point
        max_distance: maximum distance to search for the nearest neighbor

    Returns:
        A shapely Point if one is within max_distance, None otherwise

    From shapely_tools:
        @author: Dirk Eilander (dirk.eilander@deltares.nl)
        Adjusted 15-10-2019: Frederique de Groen (frederique.degroen@deltares.nl)
        Build on library from https://github.com/ojdo/python-tools/blob/master/shapelytools.py
    """

    # the point from where you are searching
    if isinstance(max_distance, pd.Series):
        max_distance = max_distance[0]
    geometry_buffered = point.buffer(max_distance)

    # expand bounds by max_distance in all directions
    # bounds = [
    #     a + b * max_distance for a, b in zip(geometry_buffered.bounds, [-1, -1, 1, 1])
    # ]

    # get list of fids where bounding boxes intersect
    interesting_points = [
        int(i) for i in spatial_index.intersection(geometry_buffered.bounds)
    ]

    if not interesting_points:
        closest_point = None
    elif len(interesting_points) == 1:
        closest_point = search_points[interesting_points[0]]
    else:
        points_list = [search_points[ip] for ip in interesting_points]
        distance_list = [
            (p, point.distance(p)) for p in points_list if point.distance(p) > 0
        ]
        closest_point, closest_distance = min(distance_list, key=lambda t: t[1])

    return closest_point


def vertices_from_lines(
    lines: list[BaseMultipartGeometry], list_ids: list[str]
) -> dict:
    """Return dict of with values: unique vertices from list of LineStrings.
    keys: index of LineString in original list
    From shapely_tools:
        Build on library from https://github.com/ojdo/python-tools/blob/master/shapelytools.py
    """
    vertices_dict = {}
    for i, line in zip(list_ids, lines):
        if isinstance(line, LineString):
            vertices_dict[i] = [Point(p) for p in set(list(line.coords))]
        if isinstance(line, MultiLineString):
            all_vertices = []
            for each_line in line.geoms:
                all_vertices.extend([Point(p) for p in set(list(each_line.coords))])
            vertices_dict[i] = all_vertices
    return vertices_dict


def create_nodes(merged_lines, crs_, cut_at_intersections):
    """Creates shapely points on intersections and endpoints of a list of shapely lines
    Args:
        merged_lines [list of shapely LineStrings]: the edges of a graph
    Returns:
        nodes [list of shapely Points]: the nodes of a graph
    """
    logging.info("Started creating nodes...")
    list_lines = list(merged_lines["geometry"])

    # Get all endpoints of all linestrings
    endpts = [
        [Point(list(line.coords)[0]), Point(list(line.coords)[-1])]
        for line in list_lines
        if isinstance(line, LineString)
    ]
    # logging.info('{} Endpoints selected'.format(len(endpts)))
    # flatten the resulting list to a simple list of points
    endpts = [pt for sublist in endpts for pt in sublist]

    more_endpts = [
        [[Point(list(ln.coords)[0]), Point(list(ln.coords)[-1])] for ln in line]
        for line in list_lines
        if isinstance(line, MultiLineString)
    ]
    more_endpts = [pt for sublist in more_endpts for pt in sublist]
    more_endpts = [pt for sublist in more_endpts for pt in sublist]
    endpts.extend(more_endpts)

    if cut_at_intersections is not True:
        # create nodes on intersections and on lines that should be snapped
        inters = []
        for line1, line2 in itertools.combinations(list_lines, 2):
            # Make points at intersections of lines
            if line1.intersects(line2):
                inter = line1.intersection(line2)
                if "Point" == inter.type:
                    inters.append(inter)
                elif "MultiPoint" == inter.type:
                    inters.extend([pt for pt in inter])
                elif "MultiLineString" == inter.type:
                    multi_line = [line for line in inter]
                    first_coords = multi_line[0].coords[0]
                    last_coords = multi_line[len(multi_line) - 1].coords[1]
                    inters.append(Point(first_coords[0], first_coords[1]))
                    inters.append(Point(last_coords[0], last_coords[1]))
                elif "GeometryCollection" == inter.type:
                    for geom in inter:
                        if "Point" == geom.type:
                            inters.append(geom)
                        elif "MultiPoint" == geom.type:
                            inters.extend([pt for pt in geom])
                        elif "MultiLineString" == geom.type:
                            multi_line = [line for line in geom]
                            first_coords = multi_line[0].coords[0]
                            last_coords = multi_line[len(multi_line) - 1].coords[1]
                            inters.append(Point(first_coords[0], first_coords[1]))
                            inters.append(Point(last_coords[0], last_coords[1]))

        # Extend the endpoints with the intersection points that are not endpoints
        endpts.extend([pt for pt in inters if pt not in endpts])

    # Add to GeoDataFrame and delete duplicate points
    unique_points = delete_duplicates(endpts)
    points_gdf = gpd.GeoDataFrame(
        {"node_fid": range(len(unique_points)), "geometry": unique_points},
        geometry="geometry",
        crs=crs_,
    )

    return points_gdf


def cut_lines(lines_gdf, nodes, id_name: str, tolerance, crs_):
    """Cuts lines at the nodes, with a certain tolerance
    Args:
        lines_gdf (geodataframe): the network with edges that should be cut
        nodes (geodataframe): points to use for cutting the edges
        idName (string): name of the Unique ID column in the lines_gdf
        tolerance: how far a point should be from the edge to cut the edge
        crs_: the CRS of the project

    Returns:
        lines_gdf (geodataframe): the network with cut edges. The IDs of the new edges counting +1 on the maximum ID number
    """
    max_id = max(lines_gdf[id_name])
    list_columns = list(lines_gdf.columns.values)
    for rem in ["geometry", "length", id_name]:
        list_columns.remove(rem)

    to_add = []
    to_remove = []
    to_iterate = zip(
        list(lines_gdf.index.values),
        list(lines_gdf[id_name]),
        list(lines_gdf["geometry"]),
    )

    for idx, i, line in to_iterate:
        if isinstance(line, LineString):
            points_to_cut = [
                pnt
                for pnt in list(nodes["geometry"])
                if (line.distance(pnt) < tolerance)
                & (line.boundary.distance(pnt) > tolerance)
            ]
        elif isinstance(line, MultiLineString):
            points_to_cut = []
            for ln in line:
                points_to_cut.extend(
                    [
                        pnt
                        for pnt in list(nodes["geometry"])
                        if (ln.distance(pnt) < tolerance)
                        & (ln.boundary.distance(pnt) > tolerance)
                    ]
                )

        if points_to_cut:
            # cut lines
            newlines = split_line_with_points(line=line, points=points_to_cut)

            # copy and remove the row of the original linestring
            properties_dict = {}
            if list_columns:
                properties_dict = lines_gdf.loc[lines_gdf[id_name] == i][
                    list_columns
                ].to_dict(orient="records")[0]

            for j, newline in enumerate(newlines):
                if j == 0:
                    # add the data with one part of the cut linestring
                    properties_dict.update(
                        {
                            id_name: i,
                            "geometry": newline,
                            "length": line_length(newline, crs_),
                        }
                    )
                    logging.info(
                        "cut line segment {} {}, added new line segment with {} {}".format(
                            id_name, i, id_name, i
                        )
                    )
                else:
                    properties_dict.update(
                        {
                            id_name: max_id + 1,
                            "geometry": newline,
                            "length": line_length(newline, crs_),
                        }
                    )
                    logging.info(
                        "cut line segment {} {}, added new line segment with {} {}".format(
                            id_name, i, id_name, properties_dict[id_name]
                        )
                    )
                    max_id += 1

                to_add.append(properties_dict)

            # remove the original linestring that has been cut
            to_remove.append(idx)

    lines_gdf.drop(to_remove, inplace=True)
    lines_gdf = lines_gdf.append(to_add, ignore_index=True)
    return lines_gdf


def split_line_with_points(line, points):
    """Splits a line string in several segments considering a list of points."""
    segments = []
    current_line = line

    # make a list of points and its distance to the start to sort them from small to large distance
    list_dist = [current_line.project(pnt) for pnt in points]
    list_dist.sort()

    for i, d in enumerate(list_dist):
        # Subtract the previous distance from the current distance to cut the segments in the right way.
        if i > 0:
            d = d - list_dist[i - 1]

        # cut the line at a distance d
        seg, current_line = cut(current_line, d)
        if seg:
            segments.append(seg)

    segments.append(current_line)
    return segments


def cut(line: BaseMultipartGeometry, distance: float) -> tuple[LineString, LineString]:
    # Cuts a line in two at a distance from its starting point
    # This is taken from shapely manual
    if (distance <= 0.0) | (distance >= line.length):
        return [None, LineString(line)]

    if isinstance(line, LineString):
        coords = list(line.coords)
        for i, p in enumerate(coords):
            pd = line.project(Point(p))
            if pd == distance:
                return [LineString(coords[: i + 1]), LineString(coords[i:])]
            if pd > distance:
                cp = line.interpolate(distance)
                # check if the LineString contains an Z-value, if so, remove
                # only use XY because otherwise the snapping functionality doesn't work
                return [
                    LineString([xy[0:2] for xy in coords[:i]] + [(cp.x, cp.y)]),
                    LineString([(cp.x, cp.y)] + [xy[0:2] for xy in coords[i:]]),
                ]
    elif isinstance(line, MultiLineString):
        for ln in line.geoms:
            coords = list(ln.coords)
            for i, p in enumerate(coords):
                pd = ln.project(Point(p))
                if pd == distance:
                    return [LineString(coords[: i + 1]), LineString(coords[i:])]
                if pd > distance:
                    cp = ln.interpolate(distance)
                    # check if the LineString contains an Z-value, if so, remove
                    # only use XY because otherwise the snapping functionality doesn't work
                    return [
                        LineString([xy[0:2] for xy in coords[:i]] + [(cp.x, cp.y)]),
                        LineString([(cp.x, cp.y)] + [xy[0:2] for xy in coords[i:]]),
                    ]


def join_nodes_edges(
    gdf_nodes: gpd.GeoDataFrame, gdf_edges: gpd.GeoDataFrame, id_name: str
) -> gpd.GeoDataFrame:
    """Creates tuples from the adjacent nodes and add as column in geodataframe.
    Args:
        gdf_nodes [geodataframe]: geodataframe of the nodes of a graph
        gdf_edges [geodataframe]: geodataframe of the edges of a graph
    Returns:
        result [geodataframe]: geodataframe of adjacent nodes from edges
    """
    logging.info("Started joining edges and nodes...")
    # list of the edges that are not topographically correct
    incorrect_edges = []

    # add node attributes to edges
    gdf = gpd.sjoin(gdf_edges, gdf_nodes, how="left", op="intersects")

    tuples_df = pd.DataFrame({"node_A": [], "node_B": []})

    for edge in gdf[id_name].unique():
        node_tuple = gdf.loc[gdf[id_name] == edge, "node_fid"]
        if len(node_tuple) > 2:
            # if there are more than 2 nodes intersecting the linestring, choose the ones at the endpoints
            # todo: check this section carefully!!
            incorrect_edges.append(edge)
            line_nodes = gdf.loc[gdf[id_name] == edge, "geometry"].iloc[0]
            if isinstance(line_nodes, LineString):
                point_coords = [
                    Point(line_nodes.coords[0]),
                    Point(line_nodes.coords[-1]),
                ]  # these are the two endpoints of the linestring - we take these as nodes
                n = gdf_nodes[
                    gdf_nodes["node_fid"].isin(node_tuple)
                ]  # get the geometries of the nodes
                special_tuple = ()
                for point in list(n.geometry):
                    if any(p.equals(point) for p in point_coords):
                        special_tuple = special_tuple + (
                            n.loc[n.geometry == point, "node_fid"].iloc[0],
                        )  # find the node id of the two endpoints of the linestring
                warnings.warn(
                    "More than two nodes are intersecting with edge {}: {}. The nodes that are intersecting are: {}".format(
                        id_name, edge, list(n["node_fid"])
                    )
                )
                try:
                    tuples_df = tuples_df.append(
                        {"node_A": special_tuple[0], "node_B": special_tuple[1]},
                        ignore_index=True,
                    )
                except IndexError as e:
                    warnings.warn(
                        "Only one node can be found for edge with {} {}: {}".format(
                            id_name, edge, e
                        )
                    )
            elif isinstance(line_nodes, MultiLineString):
                special_tuple = ()
                for ln in line_nodes:
                    point_coords = [
                        Point(ln.coords[0]),
                        Point(ln.coords[-1]),
                    ]  # these are the two endpoints of the linestring - we take these as nodes
                    n = gdf_nodes[
                        gdf_nodes["node_fid"].isin(node_tuple)
                    ]  # get the geometries of the nodes
                    for point in list(n.geometry):
                        if any(p.equals(point) for p in point_coords):
                            special_tuple = special_tuple + (
                                n.loc[n.geometry == point, "node_fid"].iloc[0],
                            )  # find the node id of the two endpoints of the linestring
                    logging.warning(
                        "More than two nodes are intersecting with edge {}: {}. The nodes that are intersecting are: {}".format(
                            id_name, edge, list(n["node_fid"])
                        )
                    )
                try:
                    tuples_df = tuples_df.append(
                        {"node_A": special_tuple[0], "node_B": special_tuple[1]},
                        ignore_index=True,
                    )
                except IndexError as e:
                    warnings.warn(
                        "Only one node can be found for edge with {} {}: {}".format(
                            id_name, edge, e
                        )
                    )
        elif len(node_tuple) < 2:
            # somehow the geopandas sjoin did not find any nodes on this edge, but there are so look for them
            node_a = [
                i
                for i, xy in zip(gdf_nodes.node_fid, gdf_nodes.geometry)
                if xy.almost_equals(
                    Point(
                        list(
                            gdf_edges.loc[gdf_edges[id_name] == edge]
                            .iloc[0]
                            .geometry.coords
                        )[0]
                    )
                )
            ]
            node_b = [
                i
                for i, xy in zip(gdf_nodes.node_fid, gdf_nodes.geometry)
                if xy.almost_equals(
                    Point(
                        list(
                            gdf_edges.loc[gdf_edges[id_name] == edge]
                            .iloc[0]
                            .geometry.coords
                        )[-1]
                    )
                )
            ]
            tuples_df = pd.concat(
                [
                    tuples_df,
                    pd.DataFrame.from_records(
                        [
                            {
                                "node_A": gdf_nodes.loc[
                                    gdf_nodes["node_fid"] == node_a[0], "node_fid"
                                ].iloc[0],
                                "node_B": gdf_nodes.loc[
                                    gdf_nodes["node_fid"] == node_b[0], "node_fid"
                                ].iloc[0],
                            },
                        ]
                    ),
                ],
                ignore_index=True,
            )

        elif len(node_tuple) == 2:
            # this is what you want for a good network
            tuples_df = pd.concat(
                [
                    tuples_df,
                    pd.DataFrame.from_records(
                        [
                            {
                                "node_A": node_tuple.iloc[0],
                                "node_B": node_tuple.iloc[1],
                            },
                        ]
                    ),
                ],
                ignore_index=True,
            )
        else:
            warnings.warn("Something went wrong..")

    if incorrect_edges:
        warnings.warn("More than 2 nodes intersecting edges {}".format(incorrect_edges))

    # reset indices in case they are not unique
    gdf_edges.reset_index(inplace=True, drop=True)
    tuples_df.reset_index(inplace=True, drop=True)

    result = gpd.GeoDataFrame(pd.concat([gdf_edges, tuples_df], axis=1))

    # drop all columns without values
    if result.columns[result.isnull().all()].any():
        to_drop = result.columns[result.isnull().all()]
        result.drop(to_drop, axis=1, inplace=True)

    logging.info("Function [join_nodes_edges]: executed")

    return result


def hazard_join_id_shp(roads, hazard_data_dict: dict):
    # read and join hazard data
    col_id, col_val = hazard_data_dict["ID"], hazard_data_dict["attribute_name"][0]

    # Fiona is not always loading the geodataframe with all data, so try a few times to get it correct
    attempts = 0
    while attempts < 3:
        try:
            hazard = gpd.read_file(hazard_data_dict["path"][0], engine="pyogrio")
            hazard = hazard[[col_id, col_val]]
            break
        except KeyError:
            attempts += 1
            print(
                "Attempt {} to load hazard data: {}".format(
                    attempts, hazard_data_dict["path"][0].split("\\")[-1]
                )
            )

    for i in range(1, len(hazard_data_dict["path"])):
        attempts = 0
        while attempts < 3:
            try:
                hazard2 = gpd.read_file(
                    hazard_data_dict["path"][i], encoding="utf-8", engine="pyogrio"
                )
                hazard = pd.concat(
                    [hazard, hazard2[[col_id, col_val]]], ignore_index=True
                )
                break
            except KeyError:
                attempts += 1
                print(
                    "Attempt {} to load hazard data: {}".format(
                        attempts, hazard_data_dict["path"][i].split("\\")[-1]
                    )
                )

    if (col_val in roads) and (col_id in roads):
        hazard = pd.concat([hazard, roads[[col_id, col_val]]], ignore_index=True)

    # Not necessary now
    # hazard.drop_duplicates(inplace=True)

    for ii in hazard[col_id].unique():
        roads.loc[roads[col_id] == ii, "_{}".format(col_val)] = max(
            hazard.loc[(hazard[col_id] == ii), col_val]
        )

    return roads


def delete_duplicates(all_points: list[Point]) -> list[Point]:
    """
    Delete duplicate points given they are 'almost' equals.

    Args:
        all_points (list[Point]): List with potentially repeated points.

    Returns:
        list[Point]: list with unique points.
    """
    points = [point for point in all_points]
    uniquepoints = []
    for point in points:
        if not any(p.almost_equals(point) for p in uniquepoints):
            uniquepoints.append(point)
    return uniquepoints


def gdf_check_create_unique_ids(
    gdf: gpd.GeoDataFrame, id_name: str, new_id_name: str = "rfid"
) -> tuple[gpd.GeoDataFrame, str]:
    """
    Check if the ID's are unique per edge: if not, add an own ID called 'fid'

    Args:
        gdf (gpd.GeoDataFrame): Dataframe with edges to check.
        id_name (str): ID to search for uniqueness.
        new_id_name (str, optional): Optinal new id name to give. Defaults to "rfid".

    Returns:
        tuple[gpd.GeoDataFrame, str]: Resulting dataframe and its ID.
    """
    check = list(gdf.index)
    logging.info("Started creating unique ids...")
    if len(gdf[id_name].unique()) == len(check):
        logging.info("Using the user-defined identifier field {}.".format(id_name))
        return gdf, id_name

    gdf[new_id_name] = check
    logging.warning(
        "Added a new unique identifier field {} because the original field '{}' "
        "did not contain unique values per road segment."
        "For further network processing, change the 'file_id' parameter in the network.ini file"
        "to '{}".format(new_id_name, id_name, new_id_name)
    )
    return gdf, new_id_name


def graph_check_create_unique_ids(
    graph: nx.Graph, id_name: str, new_id_name: str = "rfid"
) -> tuple[nx.Graph, str]:
    """
    TODO: This is not really being used. It could be removed.
    Check if the ID's are unique per edge: if not, add an own ID called 'fid'

    Args:
        graph (Graph): Graph to prune from repeated ids.
        id_name (str): ID to search.
        new_id_name (str, optional): Optional new id to set for repeated elements. Defaults to "rfid".

    Returns:
        tuple[Graph, str]: Resulting graph and used ID.
    """
    if len(set([str(e[-1][id_name]) for e in graph.edges.data(keys=True)])) < len(
        graph.edges()
    ):
        i = 0
        for u, v, k in graph.edges(data=True):
            graph[u][v][k][new_id_name] = i
            i += 1
        logging.info(
            "Added a new unique identifier field {} because the original field '{}' did not contain unique values per road segment.".format(
                new_id_name, id_name
            )
        )
        return graph, new_id_name

    return graph, id_name


def add_missing_geoms_graph(graph: nx.Graph, geom_name: str = "geometry") -> nx.Graph:
    # Not all nodes have geometry attributed (some only x and y coordinates) so add a geometry columns
    nodes_without_geom = [
        n[0] for n in graph.nodes(data=True) if geom_name not in n[-1]
    ]
    for nd in nodes_without_geom:
        graph.nodes[nd][geom_name] = Point(graph.nodes[nd]["x"], graph.nodes[nd]["y"])

    edges_without_geom = [
        e for e in graph.edges.data(keys=True, data=True) if geom_name not in e[-1]
    ]
    for ed in edges_without_geom:
        if graph.nodes[ed[0]].get(geom_name, None) and graph.nodes[ed[1]].get(
            geom_name, None
        ):
            graph[ed[0]][ed[1]][ed[2]][geom_name] = LineString(
                [graph.nodes[ed[0]][geom_name], graph.nodes[ed[1]][geom_name]]
            )
    return graph


def add_x_y_to_nodes(graph: nx.Graph) -> nx.Graph:
    """
    Add missing x and y attributes to nodes.
    TODO: Should this be moved to `network_simplification`?

    Args:
        graph (nx.Graph): Graph to add x and y attributes to

    Returns:
        nx.Graph: Graph with x and y attributes added
    """
    for _, data in graph.nodes(data=True):
        if "x" not in data or "y" not in data:
            # Use 'geometry' or provide default values if it's not present
            geometry = data.get("geometry", (0.0, 0.0))
            data.setdefault("x", round(geometry.x, 7))
            data.setdefault("y", round(geometry.y, 7))
    return graph


def graph_from_gdf(
    gdf: gpd.GeoDataFrame, gdf_nodes, name: str = "network", node_id: str = "ID"
) -> nx.MultiGraph:
    # create a Graph object
    _created_graph = nx.MultiGraph(crs=gdf.crs)

    # create nodes on the Graph
    for _, row in gdf_nodes.iterrows():
        c = {node_id: row[node_id], "geometry": row.geometry}
        _created_graph.add_node(row[node_id], **c)

    # create edges on top of the nodes
    for _, row in gdf.iterrows():
        dict_row = row.to_dict()
        _created_graph.add_edge(
            u_for_edge=dict_row["node_A"], v_for_edge=dict_row["node_B"], **dict_row
        )

    # make a name
    _created_graph.graph["name"] = name

    return _created_graph


def graph_to_gdf(
    graph_to_convert: nx.Graph,
    save_nodes: bool = False,
    save_edges: bool = True,
    to_save: bool = False,
):
    """Takes in a networkx graph object and returns edges and nodes as geodataframes
    Arguments:
        graph_to_convert (Graph): networkx graph object to be converted
        save_nodes (bool): get the nodes as a geodataframe (False)
        save_edges (bool): get the edges as a geodataframe (True)
        to_save (bool): convert all columns to string (False)

    Returns:
        edges (GeoDataFrame): contains the edges
        nodes (GeoDataFrame): contains the nodes
    """

    nodes, edges = None, None
    if save_nodes and save_edges:
        nodes, edges = graph_to_gdfs(
            graph_to_convert, nodes=save_nodes, edges=save_edges, node_geometry=False
        )
        if to_save:
            for df in [edges, nodes]:
                for col in df.columns:
                    if df[col].dtype == object and col != df.geometry.name:
                        df[col] = df[col].astype(str)
    elif not save_nodes and save_edges:
        edges = graph_to_gdfs(graph_to_convert, nodes=save_nodes, edges=save_edges)
    elif save_nodes and not save_edges:
        nodes = graph_to_gdfs(graph_to_convert, nodes=save_nodes, edges=save_edges)

    return edges, nodes


def graph_to_gpkg(origin_graph: nx.Graph, edge_gpkg: str, node_gpkg: str) -> None:
    """Takes in a networkx graph object and outputs shapefiles at the paths indicated by edge_gpkg and node_gpkg

    Arguments:
        origin_graph [nx.Graph]: networkx graph object to be converted
        edge_gpkg [str]: output path including extension for edges geopackage
        node_gpkg [str]: output path including extension for nodes geopackage

    Returns:
        None
    """
    # now only multidigraphs and graphs are used
    if type(origin_graph) == nx.Graph:
        origin_graph = nx.MultiGraph(origin_graph)

    # The nodes should have a geometry attribute (perhaps on top of the x and y attributes)
    nodes, edges = graph_to_gdfs(origin_graph, node_geometry=False)

    dfs = [edges, nodes]
    for df in dfs:
        for col in df.columns:
            if df[col].dtype == object and col != df.geometry.name:
                df[col] = df[col].astype(str)

    # Add a CRS to the nodes
    if nodes.crs is None and edges.crs is not None:
        nodes.crs = edges.crs

    logging.info("Saving nodes as shapefile: {}".format(node_gpkg))
    logging.info("Saving edges as shapefile: {}".format(edge_gpkg))

    # The encoding utf-8 might result in an empty shapefile if the wrong encoding is used.
    for entity in [nodes, edges]:
        if "osmid" in entity:
            # Otherwise it gives this error: cannot insert osmid, already exist
            entity["osmid_original"] = entity.pop("osmid")
    for _path in [node_gpkg, edge_gpkg]:
        if _path.exists():
            _path.unlink()
    nodes.to_file(node_gpkg, driver="GPKG", encoding="utf-8")
    edges.to_file(edge_gpkg, driver="GPKG", encoding="utf-8")


@staticmethod
def get_normalized_geojson_polygon(geojson_path: Path) -> BaseGeometry:
    """
    Converts a GeoJson object to a Shapely Polygon.

    Args:
        geojson_path (Path): File containing a geojson object.

    Raises:
        ValueError: When the polygon longitude is out of bounds.
        ValueError: When the polygon latitude is out of bounds.

    Returns:
        BaseGeometry: Resulting normalized (in lat and lon) polygon in shapely format.
    """
    _read_geojson = gpd.read_file(geojson_path)

    def check_bounds(geometry: BaseGeometry):
        if geometry.bounds[0] > 180 or geometry.bounds[2] < -180:
            raise ValueError(
                "Longitude is out of bounds, check your JSON format or data. The Coordinate Reference System should be in EPSG:4326."
            )
        if geometry.bounds[1] > 90 or geometry.bounds[3] < -90:
            raise ValueError(
                "Latitude is out of bounds, check your JSON format or data. The Coordinate Reference System should be in EPSG:4326."
            )

    # Discard z-coordinate, if it exists
    check_bounds(_read_geojson.geometry[0])

    # Create a shapely polygon from the coordinates.
    return shape(_read_geojson.geometry[0]).buffer(0)


def read_merge_shp(shp_file_analyse, id_name, shp_file_diversion=[], crs_=4326):
    """Imports shapefile(s) and saves attributes in a pandas dataframe.

    Args:
        shp_file_analyse (string or list of strings): absolute path(s) to the shapefile(s) that will be used for analysis
        shp_file_diversion (string or list of strings): absolute path(s) to the shapefile(s) that will be used to calculate alternative routes but is not analysed
        id_name (string): the name of the Unique ID column
        crs_ (int): the EPSG number of the coordinate reference system that is used
    Returns:
        lines (list of shapely LineStrings): full list of linestrings
        properties (pandas dataframe): attributes of shapefile(s), in order of the linestrings in lines
    """

    # convert shapefile names to a list if it was not already a list
    if isinstance(shp_file_analyse, str):
        shp_file_analyse = [shp_file_analyse]
    if isinstance(shp_file_diversion, str):
        shp_file_diversion = [shp_file_diversion]

    lines = []

    # read the shapefile(s) for analysis
    for shp in shp_file_analyse:
        lines_shp = gpd.read_file(shp, engine="pyogrio")
        lines_shp["to_analyse"] = 1
        lines.append(lines_shp)

    # read the shapefile(s) for only diversion
    if isinstance(shp_file_diversion, list):
        for shp2 in shp_file_diversion:
            lines_shp = gpd.read_file(shp2, engine="pyogrio")
            lines_shp["to_analyse"] = 0
            lines.append(lines_shp)

    # concatenate all shapefiles into one geodataframe
    lines = pd.concat(lines)
    lines.crs = crs_

    # append the length of the road stretches
    lines["length"] = lines["geometry"].apply(lambda x: line_length(x, crs_))

    if lines["geometry"].apply(lambda row: isinstance(row, MultiLineString)).any():
        for line in lines.loc[
            lines["geometry"].apply(lambda row: isinstance(row, MultiLineString))
        ].iterrows():
            if len(linemerge(line[1].geometry)) > 1:
                warnings.warn(
                    "Edge with {} = {} is a MultiLineString, which cannot be merged to one line. Check this part.".format(
                        id_name, line[1][id_name]
                    )
                )

    print(
        "Shapefile(s) loaded with attributes: {}.".format(list(lines.columns.values))
    )  # fill in parameter names

    return lines


def check_hazard_extent_resolution(list_hazards):
    """Checks whether the extent of a list of hazard rasters is exactly the same.

    Args:
        list_hazards (list of pathlib paths or strings): A list of paths to the hazard map rasters of which
            the extent needs to be checked

    Returns:
        (bool): True if the extents are exactly the same, False if they are not.
    """
    if len(list_hazards) == 1:
        return True
    check_hazard_extent = [
        gdal.Open(str(haz)).GetGeoTransform() for haz in list_hazards
    ]
    if len(set(check_hazard_extent)) == 1:
        # All hazard have the exact same extents and resolution
        return True
    else:
        return False


def get_extent(dataset):
    cols = dataset.RasterXSize
    rows = dataset.RasterYSize
    transform = dataset.GetGeoTransform()
    minx = transform[0]
    maxx = transform[0] + cols * transform[1] + rows * transform[2]

    miny = transform[3] + cols * transform[4] + rows * transform[5]
    maxy = transform[3]

    width = maxx - minx
    height = maxy - miny

    return {
        "minX": minx,
        "maxX": maxx,
        "minY": miny,
        "maxY": maxy,
        "cols": cols,
        "rows": rows,
        "width": width,
        "height": height,
        "pixelWidth": transform[1],
        "pixelHeight": transform[5],
    }


def get_graph_edges_extent(
    network_graph: nx.Graph,
) -> tuple[float, float, float, float]:
    """Inspects all geometries of the edges of a graph and returns the most extreme coordinates

    Arguments:
        network_graph (networkX) : NetworkX graph, should have an attribute 'geometry' containty shapely geometries

    Returns
        extent (tuple) : (minX,maxX,minY,maxY)
    """
    # Start with the bounds of the (random) first edge
    (min_x, min_y, max_x, max_y) = list(network_graph.edges.data("geometry"))[0][
        -1
    ].bounds

    # Compare with the bounds of all other edges to see if there are linestrings with more extreme bounds
    for _, _, geom in network_graph.edges.data("geometry"):
        # print(u, v, geom)
        (
            minx,
            miny,
            maxx,
            maxy,
        ) = geom.bounds  # shapely returns (minx, miny, maxx, maxy)
        if minx < min_x:
            min_x = minx
        if maxx > max_x:
            max_x = maxx
        if miny < min_y:
            min_y = miny
        if maxy > max_y:
            max_y = maxy

    return min_x, max_x, min_y, max_y


def reproject_graph(original_graph: nx.Graph, crs_in: str, crs_out: str) -> nx.Graph:
    """
    Reprojects the shapely geometry data (of a NetworkX graph) to a different projection

    Arguments:
        original_graph (NetworkX graph): needs a geometry attribute containing shapely linestrings
        crs_in (string): input projection, follow the Geopandas crs convention: "espg:3035" or "EPSG4326"
        crs_out (string): output projection to convert to, idem.

    Returns:
        G_out (NetworkX graph)
    """
    att = nx.get_edge_attributes(original_graph, "geometry")
    df = pd.DataFrame.from_dict(data=att, orient="index", columns=["geometry"])
    gdf = gpd.GeoDataFrame(df)
    gdf = gdf.set_crs(crs=crs_in)
    gdf_out = gdf.to_crs(crs=crs_out)
    _reprojected_graph = original_graph.copy()
    set_values = gdf_out.to_dict(orient="index")
    nx.set_edge_attributes(_reprojected_graph, values=set_values)
    return _reprojected_graph


def bounds_intersect_1d(tup1, tup2):
    """Check if bounds of A and B intersect anywhere (have any overlapping points)

    Arguments:
        tup1 (tuple) : (min, max)
        tup2 (tuple) : (min, max)

    Returns:
        Boolean (True/False)
    """
    return (tup1[0] <= tup2[1]) and (tup1[1] >= tup2[0])


def bounds_intersect_2d(extent1, extent2):
    """
    check if bounds of A and B intersect anywhere (have any overlapping area)

    Arguments:
        *extent1* (tuple) : (minx,maxx,miny,maxy)
        *extent2* (tuple) : (minx,maxx,miny,maxy)

    Returns:
        Boolean (True/False)
    """
    x_overlap = bounds_intersect_1d((extent1[0], extent1[1]), (extent2[0], extent2[1]))
    y_overlap = bounds_intersect_1d((extent1[2], extent1[3]), (extent2[2], extent2[3]))
    return x_overlap and y_overlap


def convert_osm(osm_convert_path, pbf, o5m):
    """Converse an osm PBF file to o5m"""
    command = '""{}"  "{}" --complete-ways --drop-broken-refs -o="{}""'.format(
        osm_convert_path, pbf, o5m
    )
    os.system(command)


def filter_osm(osm_filter_path, o5m, filtered_o5m, tags=None):
    """Filters an o5m OSM file to only motorways, trunks, primary and secondary roads (or tags, if specified)"""
    if tags is None:
        tags = [
            "motorway",
            "motorway_link",
            "primary",
            "primary_link",
            "secondary",
            "secondary_link",
            "trunk",
            "trunk_link",
        ]
    command = '""{}"  "{}" --keep="highway={}" > "{}""'.format(
        osm_filter_path, o5m, " =".join(tags), filtered_o5m
    )
    os.system(command)


def graph_link_simple_id_to_complex(
    graph_simple: nx.Graph, new_id: str
) -> tuple[dict, dict]:
    """
    Create lookup tables (dicts) to match edges_ids of the complex and simple graph
    Optionally, saves these lookup tables as json files.

    Arguments:
        graph_simple (Graph) : Graph, containing attribute 'new_id'
        new_id (string) : Name of the ID attribute in graph_simple

    Returns:
        simple_to_complex (dict): Keys are ids of the simple graph, values are lists with all matching complex ids
        complex_to_simple (dict): Keys are the ids of the complex graph, value is the matching simple_ID

    We need this because the simple graph is derived from the complex graph, and therefore initially only the
    simple graph knows from which complex edges it was created. To assign this information also to the complex
    graph we invert the look-up dictionary
    @author: Kees van Ginkel en Margreet van Marle
    """
    # Iterate over the simple, because this already has the corresponding complex information
    lookup_dict = {}
    # keys are the ids of the simple graph, values are lists with all matching complex id's
    for u, v, k in tqdm(graph_simple.edges(keys=True)):
        key_1 = graph_simple[u][v][k]["{}".format(new_id)]
        value_1 = graph_simple[u][v][k]["{}_c".format(new_id)]
        lookup_dict[key_1] = value_1

    inverted_lookup_dict = {}
    # keys are the ids of the complex graph, value is the matching simple_ID
    for key, value in lookup_dict.items():
        if isinstance(value, list):
            for subvalue in value:
                inverted_lookup_dict[subvalue] = key
        elif isinstance(value, int):
            inverted_lookup_dict[value] = key

    simple_to_complex = lookup_dict
    complex_to_simple = inverted_lookup_dict

    logging.info("Lookup tables from complex to simple and vice versa were created")
    return simple_to_complex, complex_to_simple


def add_simple_id_to_graph_complex(
    complex_graph: nx.Graph, complex_to_simple, new_id
) -> nx.Graph:
    """Adds the appropriate ID of the simple graph to each edge of the complex graph as a new attribute 'rfid'

    Arguments:
        complex_graph (Graph) : The complex graph, still lacking 'rfid'
        complex_to_simple (dict) : lookup table linking complex to simple graphs

    Returns:
        complex_graph (Graph) : Same object, with added attribute 'rfid'

    """

    obtained_complex_ids = nx.get_edge_attributes(
        complex_graph, "{}_c".format(new_id)
    )  # {(u,v,k) : 'rfid_c'}
    simple_ids_per_complex_id = obtained_complex_ids  # start with a copy

    for key, value in obtained_complex_ids.items():  # {(u,v,k) : 'rfid_c'}
        try:
            new_value = complex_to_simple[
                value
            ]  # find simple id belonging to the complex id
            simple_ids_per_complex_id[key] = new_value
        except KeyError as e:
            logging.error(
                "Could not find the simple ID belonging to complex ID {}; value set to None. Full error: {}".format(
                    key, e
                )
            )
            simple_ids_per_complex_id[key] = None

    # Now the format of simple_ids_per_complex_id is: {(u,v,k) : 'rfid}
    nx.set_edge_attributes(complex_graph, simple_ids_per_complex_id, new_id)

    return complex_graph


def fraction_flooded(line: LineString, hazard_map: str):
    """Calculates the fraction of a linestring that overlaps with a hazard raster with value > 0

    Args:
        line (LineString): A single linestring that should be overlayed with the hazard map.
        hazard_map (string): Full path to the hazard map.

    Returns:
        (float) The fraction of the linestring that overlaps with the hazard raster with a value > 0.
        (0) If there is a ValueError, for example, if the linestring has no overlap with the raster,
        the function returns 0.
    """

    bbox_line = box(*line.bounds)
    try:
        with rasterio.open(hazard_map) as src:
            out_image, out_transform = mask(
                src, [bbox_line], crop=True, all_touched=True
            )

        flooded_cells = unary_union(
            [
                shape(x[0])
                for x in shapes(out_image, transform=out_transform)
                if x[-1] > 0
            ]
        )
        flooded_cells = gpd.GeoDataFrame({"geometry": [flooded_cells]})

        line_intersect = flooded_cells.intersection(line)
        return line_intersect.length.sum() / line.length
    except ValueError:
        return 0
    except Exception as e:
        logging.info("fraction_flooded() {} \n for line {}".format(e, line))


def check_crs_gdf(gdf: gpd.GeoDataFrame, crs) -> None:
    if gdf.crs != crs:
        _error = (
            "Shape projection is epsg:{} - only projection epsg:{} is allowed. ".format(
                gdf.crs, crs
            )
        )
        logging.error(_error)
        raise ValueError(_error)


def clean_memory(list_delete: list) -> None:
    """
    TODO: Not sure why this method is here, is this even working?
    """
    for to_delete in list_delete:
        del to_delete


def get_valid_mean(x_value: MaskedArray, **kwargs) -> Optional[float]:
    # **kwargs should not be removed. properties var is passed in zonal_stats. So properties should not be removed,
    #  else this will not be activated
    if not isinstance(x_value, MaskedArray):
        return np.nan
    if x_value.mask.all():
        return np.nan
    return np.mean(x_value)  # You know it's a valid type, so return the mean.


def buffer_geometry(
    gdf: gpd.GeoDataFrame, buffer: float, cap_stype: int = 3
) -> gpd.GeoDataFrame:
    """buffer the geometry of the geodataframe"""
    gdf["geometry"] = gdf["geometry"].buffer(
        buffer,
        cap_style=cap_stype,
    )
    return gdf



"""
                    GNU GENERAL PUBLIC LICENSE
                      Version 3, 29 June 2007
    Risk Assessment and Adaptation for Critical Infrastructure (RA2CE).
    Copyright (C) 2023 Stichting Deltares
    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

from __future__ import annotations

import logging
from pathlib import Path
from typing import Any

import networkx as nx
import osmnx
import pyproj
from geopandas import GeoDataFrame
from networkx import MultiDiGraph, MultiGraph
from shapely.geometry.base import BaseGeometry

import ra2ce.network.networks_utils as nut
from ra2ce.network.avg_speed.avg_speed_calculator import AvgSpeedCalculator
from ra2ce.network.exporters.json_exporter import JsonExporter
from ra2ce.network.network_config_data.enums.network_type_enum import NetworkTypeEnum
from ra2ce.network.network_config_data.enums.road_type_enum import RoadTypeEnum
from ra2ce.network.network_config_data.network_config_data import NetworkConfigData
from ra2ce.network.network_simplification import NetworkGraphSimplificator
from ra2ce.network.network_wrappers.network_wrapper_protocol import (
    NetworkWrapperProtocol,
)
from ra2ce.network.network_wrappers.osm_network_wrapper.extremities_data import (
    ExtremitiesData,
)
from ra2ce.network.network_wrappers.osm_network_wrapper.osm_utils import (
    get_node_nearest_edge,
    is_endnode_check,
    modify_graph,
)


class OsmNetworkWrapper(NetworkWrapperProtocol):

    polygon_graph: MultiDiGraph
    network_type: NetworkTypeEnum
    road_types: list[RoadTypeEnum]

    def __init__(self, config_data: NetworkConfigData) -> None:
        self.attributes_to_exclude_in_simplification = (
            config_data.network.attributes_to_exclude_in_simplification
        )
        self.output_graph_dir = config_data.output_graph_dir
        self.graph_crs = config_data.crs

        # Network
        self.network_type = config_data.network.network_type
        self.road_types = config_data.network.road_types
        self.polygon_graph = self._get_clean_graph_from_osm(config_data.network.polygon)
        self.is_directed = config_data.network.directed

    @classmethod
    def with_polygon(
        cls, config_data: NetworkConfigData, polygon: BaseGeometry
    ) -> OsmNetworkWrapper:
        """
        Gets an `OsmNetworkWrapper` with the given `polygon` transformed into a
        clean graph as the `polygon_graph` property.

        Args:
            polygon (BaseGeometry): Base polygon from which to generate the graph.

        Returns:
            OsmNetworkWrapper: Wrapper with valid `polygon_graph` property.
        """
        _wrapper = cls(config_data)
        _clean_graph = _wrapper._download_clean_graph_from_osm(
            polygon=polygon,
            network_type=_wrapper.network_type,
            road_types=_wrapper.road_types,
        )
        _wrapper.polygon_graph = _clean_graph
        return _wrapper

    @staticmethod
    def get_network_from_polygon(
        config_data: NetworkConfigData, polygon: BaseGeometry
    ) -> tuple[MultiGraph, GeoDataFrame]:
        """
        Gets a valid network (`MultiGraph` and `GeoDataFrame`) for the given network configuration and
        boundary box (represented by a `shapely.BaseGeometry`).

        Args:
            config_data (NetworkConfigData): Network data configuration required for OSM download.
            polygon (BaseGeometry): Polygon representing the boundary box.

        Returns:
            tuple[MultiGraph, GeoDataFrame]: Resulting network representations.
        """
        _network_wrapper = OsmNetworkWrapper.with_polygon(config_data, polygon)
        return _network_wrapper.get_network()

    @staticmethod
    def get_network_from_geojson(config_data: NetworkConfigData):
        """
        Gets a valid network (`MultiGraph` and `GeoDataFrame`) for the given network configuration,
        given that its network section contains a valid path for the polygon property.

        Args:
            config_data (NetworkConfigData): Network data configuration required for OSM download.

        Raises:
            ValueError: When the `NetworkConfigData` does not contain a valid polygon path
            for its network section.

        Returns:
            tuple[MultiGraph, GeoDataFrame]: Resulting network representations.
        """
        if (
            not isinstance(config_data.network.polygon, Path)
            or not config_data.network.polygon.is_file()
        ):
            raise ValueError(
                "A valid network polygon (.geojson) file path needs to be provided."
            )

        return OsmNetworkWrapper(config_data).get_network()

    def get_network(self) -> tuple[MultiGraph, GeoDataFrame]:

        # Create 'graph_simple'
        graph_simple, graph_complex, link_tables = NetworkGraphSimplificator(
            graph_complex=self.polygon_graph,
            attributes_to_exclude=self.attributes_to_exclude_in_simplification,
        ).simplify()

        # Assign the average speed and time to the graphs
        graph_simple = AvgSpeedCalculator(graph_simple, self.output_graph_dir).assign()
        graph_complex = AvgSpeedCalculator(
            graph_complex, self.output_graph_dir
        ).assign()

        # Create 'edges_complex', convert complex graph to geodataframe
        logging.info("Start converting the graph to a geodataframe")
        edges_complex, _ = nut.graph_to_gdf(graph_complex)
        logging.info("Finished converting the graph to a geodataframe")

        # Save the link tables linking complex and simple IDs
        self._export_linking_tables(link_tables)

        if not self.is_directed and isinstance(graph_simple, MultiDiGraph):
            graph_simple = graph_simple.to_undirected()

        # Check if all geometries between nodes are there, if not, add them as a straight line.
        graph_simple = nut.add_missing_geoms_graph(graph_simple, geom_name="geometry")

        return graph_simple, edges_complex

    def _export_linking_tables(self, linking_tables: tuple[Any]) -> None:
        if not self.output_graph_dir:
            logging.warning(
                "No `output_graph_dir` is set, therefore no intermediate results will be exported."
            )
            return
        _exporter = JsonExporter()
        _exporter.export(
            self.output_graph_dir.joinpath("simple_to_complex.json"), linking_tables[0]
        )
        _exporter.export(
            self.output_graph_dir.joinpath("complex_to_simple.json"), linking_tables[1]
        )

    def _get_clean_graph_from_osm(self, polygon_path: Path) -> MultiDiGraph | None:
        """
        Creates a network from a polygon by downloading via the OSM API in its extent.

        Args:
            polygon_path (Path): Path where the polygon file can be found.

        Returns:
            MultiDiGraph: Complex (clean) graph after download from OSM, for use in the damages analyses and input to derive simplified network.
        """
        # It can only read in one geojson
        if not isinstance(polygon_path, Path):
            logging.warning("No valid value provided for polygon file.")
            return None
        elif not polygon_path.is_file():
            logging.error("No polygon_file file found at {}.".format(polygon_path))
            return None

        _normalized_polygon = nut.get_normalized_geojson_polygon(polygon_path)
        _complex_graph = self._download_clean_graph_from_osm(
            polygon=_normalized_polygon,
            network_type=self.network_type,
            road_types=self.road_types,
        )
        return _complex_graph

    def _download_clean_graph_from_osm(
        self,
        polygon: BaseGeometry,
        road_types: list[RoadTypeEnum],
        network_type: NetworkTypeEnum,
    ) -> MultiDiGraph:
        _available_road_types = road_types and any(road_types)
        _road_types_as_str = (
            list(map(lambda x: x.config_value, road_types))
            if _available_road_types
            else []
        )

        if not _available_road_types and not network_type:
            raise ValueError("Either of the link_type or network_type should be known")
        elif not _available_road_types:
            # The user specified only the network type.
            _complex_graph = osmnx.graph_from_polygon(
                polygon=polygon,
                network_type=network_type.config_value,
                simplify=False,
                retain_all=True,
            )
        elif not network_type:
            # The user specified only the road types.
            cf = f'["highway"~"{"|".join(_road_types_as_str)}"]'
            _complex_graph = osmnx.graph_from_polygon(
                polygon=polygon, custom_filter=cf, simplify=False, retain_all=True
            )
        else:
            # _available_road_types and network_type
            cf = f'["highway"~"{"|".join(_road_types_as_str)}"]'
            _complex_graph = osmnx.graph_from_polygon(
                polygon=polygon,
                network_type=network_type.config_value,
                custom_filter=cf,
                simplify=False,
                retain_all=True,
            )

        logging.info(
            "graph downloaded from OSM with {:,} nodes and {:,} edges".format(
                len(list(_complex_graph.nodes())), len(list(_complex_graph.edges()))
            )
        )
        if "crs" not in _complex_graph.graph.keys():
            _complex_graph.graph["crs"] = self.graph_crs
        self.get_clean_graph(_complex_graph)
        return _complex_graph

    @staticmethod
    def get_clean_graph(complex_graph: MultiDiGraph):
        complex_graph = OsmNetworkWrapper.drop_duplicates(complex_graph)
        complex_graph = nut.add_missing_geoms_graph(
            graph=complex_graph, geom_name="geometry"
        ).to_directed()
        complex_graph = OsmNetworkWrapper.snap_nodes_to_nodes(
            graph=complex_graph, threshold=0.00005
        )
        return complex_graph

    @staticmethod
    def drop_duplicates(complex_graph: MultiDiGraph) -> MultiDiGraph:
        unique_elements = (
            set()
        )  # This gets updated during the drop_duplicates_in_nodes and drop_duplicates_in_edges

        unique_graph = OsmNetworkWrapper.drop_duplicates_in_nodes(
            unique_elements=unique_elements, graph=complex_graph
        )
        unique_graph = OsmNetworkWrapper.drop_duplicates_in_edges(
            unique_elements=unique_elements,
            unique_graph=unique_graph,
            graph=complex_graph,
        )
        return unique_graph

    @staticmethod
    def drop_duplicates_in_nodes(
        unique_elements: set, graph: MultiDiGraph
    ) -> MultiDiGraph:
        if unique_elements is None or not isinstance(unique_elements, set):
            raise ValueError("unique_elements should be a set")

        unique_graph = nx.MultiDiGraph()
        for node, data in graph.nodes(data=True):
            if data["x"] is None or data["y"] is None:
                raise ValueError(
                    "Incompatible coordinate keys. Check the keys which define the x and y coordinates"
                )

            x, y = data["x"], data["y"]
            coord = (x, y)
            if coord not in unique_elements:
                node_attributes = {key: value for key, value in data.items()}
                unique_graph.add_node(node, **node_attributes)
                unique_elements.add(coord)
        # Copy the graph dictionary from the source one.
        unique_graph.graph = graph.graph
        return unique_graph

    @staticmethod
    def drop_duplicates_in_edges(
        unique_elements: set, unique_graph: MultiDiGraph, graph: MultiDiGraph
    ):
        """
        Checks if both extremities are in the unique_graph (u has not the same coor of v, no line from u to itself is
        allowed). Checks if an edge is already made between such extremities with the given id and coordinates before
        considering it in the unique graph
        """
        if (
            not unique_elements
            or not any(unique_elements)
            or not all(isinstance(item, tuple) for item in unique_elements)
        ):
            raise ValueError(
                """unique_elements cannot be None, empty, or have non-tuple elements. 
            Provide a set with all unique node coordinates as tuples of (x, y)"""
            )
        if unique_graph is None:
            raise ValueError(
                """unique_graph cannot be None. Provide a graph with unique nodes or perform the 
        drop_duplicates_in_nodes on the graph to generate a unique_graph"""
            )

        def validity_check(extremities_tuple) -> bool:
            extremities = extremities_tuple[0]
            return not (
                extremities.from_to_id is None
                or extremities.from_to_coor is None
                or extremities.from_id == extremities.to_id
            )

        def valid_extremity_data(u, v, data) -> tuple[ExtremitiesData, dict]:
            _extremities_data = ExtremitiesData.get_extremities_data_for_sub_graph(
                from_node_id=u,
                to_node_id=v,
                sub_graph=unique_graph,
                graph=graph,
                shared_elements=unique_elements,
            )

            return _extremities_data, data

        for _extremity_data, _edge_data in filter(
            validity_check,
            map(lambda edge: valid_extremity_data(*edge), graph.edges(data=True)),
        ):
            _id_combination = (_extremity_data.from_to_id, _extremity_data.to_from_id)
            _coor_combination = (
                _extremity_data.from_to_coor,
                _extremity_data.to_from_coor,
            )
            if all(
                _combination not in unique_elements
                for _combination in [_id_combination, _coor_combination]
            ):
                edge_attributes = {key: value for key, value in _edge_data.items()}
                unique_graph.add_edge(
                    _extremity_data.from_id, _extremity_data.to_id, **edge_attributes
                )
                unique_elements.add(_id_combination)
                unique_elements.add(_coor_combination)

        return unique_graph

    @staticmethod
    def snap_nodes_to_nodes(graph: MultiDiGraph, threshold: float) -> MultiDiGraph:
        return osmnx.consolidate_intersections(
            G=graph, rebuild_graph=True, tolerance=threshold, dead_ends=False
        )

    @staticmethod
    def snap_nodes_to_edges(graph: MultiDiGraph, threshold: float) -> MultiDiGraph:
        def threshold_check(node_nearest_ed: dict):
            distance = node_nearest_ed["nearest_edge"][-1]
            return distance <= threshold

        end_nodes = [
            node for node in graph.nodes(data=True) if is_endnode_check(graph, node[0])
        ]

        if (
            not graph.graph
            or not graph.graph["crs"]
            or isinstance(graph.graph["crs"], str)
        ):
            graph.graph["crs"] = pyproj.CRS("epsg:4326")
        nut.add_missing_geoms_graph(graph=graph, geom_name="geometry").to_directed()

        for node_nearest_edge_data in filter(
            threshold_check,
            map(lambda end_node: get_node_nearest_edge(graph, end_node), end_nodes),
        ):
            modify_graph(graph=graph, node_nearest_edge_data=node_nearest_edge_data)
        return graph
